import copy
import inspect
import warnings

import numpy as np
import tensorflow as tf
import tensorflow.keras.layers as layers
from notekeras import ops
from notekeras.ops import deserialize_scatter, dot, serialize_scatter
from scipy import sparse as sp
from scipy.sparse.linalg import ArpackNoConvergence
from tensorflow.keras import activations
from tensorflow.keras import backend as K
from tensorflow.keras import constraints, initializers, regularizers
from tensorflow.keras.layers import (BatchNormalization, Dense, Dropout,
                                     GRUCell, Layer, PReLU)
from tensorflow.keras.models import Sequential

from .utils import *


class MessagePassing(Layer):
    r"""
    A general class for message passing networks from the paper

    > [Neural Message Passing for Quantum Chemistry](https://arxiv.org/abs/1704.01212)<br>
    > Justin Gilmer et al.

    **Mode**: single, disjoint.

    **This layer and all of its extensions expect a sparse adjacency matrix.**

    This layer computes:
    $$
        \x_i' = \gamma \left( \x_i, \square_{j \in \mathcal{N}(i)} \,
        \phi \left(\x_i, \x_j, \e_{j \rightarrow i} \right) \right),
    $$

    where \( \gamma \) is a differentiable update function, \( \phi \) is a
    differentiable message function, \( \square \) is a permutation-invariant
    function to aggregate the messages (like the sum or the average), and
    \(\E_{ij}\) is the edge attribute of edge i-j.

    By extending this class, it is possible to create any message-passing layer
    in single/disjoint mode.

    **API**

    ```python
    propagate(x, a, e=None, **kwargs)
    ```
    Propagates the messages and computes embeddings for each node in the graph. <br>
    Any `kwargs` will be forwarded as keyword arguments to `message()`,
    `aggregate()` and `update()`.

    ```python
    message(x, **kwargs)
    ```
    Computes messages, equivalent to \(\phi\) in the definition. <br>
    Any extra keyword argument of this function will be populated by
    `propagate()` if a matching keyword is found. <br>
    Use `self.get_i()` and  `self.get_j()` to gather the elements using the
    indices `i` or `j` of the adjacency matrix. Equivalently, you can access
    the indices themselves via the `index_i` and `index_j` attributes.

    ```python
    aggregate(messages, **kwargs)
    ```
    Aggregates the messages, equivalent to \(\square\) in the definition. <br>
    The behaviour of this function can also be controlled using the `aggregate`
    keyword in the constructor of the layer (supported aggregations: sum, mean,
    max, min, prod). <br>
    Any extra keyword argument of this function will be  populated by
    `propagate()` if a matching keyword is found.

    ```python
    update(embeddings, **kwargs)
    ```
    Updates the aggregated messages to obtain the final node embeddings,
    equivalent to \(\gamma\) in the definition. <br>
    Any extra keyword argument of this function will be  populated by
    `propagate()` if a matching keyword is found.

    **Arguments**:

    - `aggregate`: string or callable, an aggregation function. This flag can be
    used to control the behaviour of `aggregate()` wihtout re-implementing it.
    Supported aggregations: 'sum', 'mean', 'max', 'min', 'prod'.
    If callable, the function must have the signature `foo(updates, indices, n_nodes)`
    and return a rank 2 tensor with shape `(n_nodes, ...)`.
    - `kwargs`: additional keyword arguments specific to Keras' Layers, like
    regularizers, initializers, constraints, etc.
    """

    def __init__(self,
                 aggregate="sum",
                 activation=None,
                 use_bias=True,
                 kernel_initializer="glorot_uniform",
                 bias_initializer="zeros",
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint

        self.bias = None

        kwargs.update({
            activation: activation,
            use_bias: use_bias,
            kernel_initializer: kernel_initializer,
            bias_initializer: bias_initializer,
            kernel_regularizer: kernel_regularizer,
            bias_regularizer: bias_regularizer,
            activity_regularizer: activity_regularizer,
            kernel_constraint: kernel_constraint,
            bias_constraint: bias_constraint,
        })
        super().__init__(
            **{k: v for k, v in kwargs.items() if is_keras_kwarg(k)})
        self.kwargs_keys = []

        for key in kwargs:
            if is_layer_kwarg(key):
                attr = kwargs[key]
                attr = deserialize_kwarg(key, attr)
                self.kwargs_keys.append(key)
                setattr(self, key, attr)

        self.msg_signature = inspect.signature(self.message).parameters
        self.agg_signature = inspect.signature(self.aggregate).parameters
        self.upd_signature = inspect.signature(self.update).parameters
        self.agg = deserialize_scatter(aggregate)

    def call(self, inputs, **kwargs):
        x, a, e = self.get_inputs(inputs)
        return self.propagate(x, a, e)

    def build(self, input_shape):
        self.built = True

    def propagate(self, x, a, e=None, **kwargs):
        self.n_nodes = tf.shape(x)[-2]
        self.index_i = a.indices[:, 1]
        self.index_j = a.indices[:, 0]

        # Message
        msg_kwargs = self.get_kwargs(x, a, e, self.msg_signature, kwargs)
        messages = self.message(x, **msg_kwargs)

        # Aggregate
        agg_kwargs = self.get_kwargs(x, a, e, self.agg_signature, kwargs)
        embeddings = self.aggregate(messages, **agg_kwargs)

        # Update
        upd_kwargs = self.get_kwargs(x, a, e, self.upd_signature, kwargs)
        output = self.update(embeddings, **upd_kwargs)

        return output

    def message(self, x, **kwargs):
        return self.get_j(x)

    def aggregate(self, messages, **kwargs):
        return self.agg(messages, self.index_i, self.n_nodes)

    def update(self, embeddings, **kwargs):
        return embeddings

    def get_i(self, x):
        return tf.gather(x, self.index_i, axis=-2)

    def get_j(self, x):
        return tf.gather(x, self.index_j, axis=-2)

    def get_kwargs(self, x, a, e, signature, kwargs):
        output = {}
        for k in signature.keys():
            if signature[k].default is inspect.Parameter.empty or k == "kwargs":
                pass
            elif k == "x":
                output[k] = x
            elif k == "a":
                output[k] = a
            elif k == "e":
                output[k] = e
            elif k in kwargs:
                output[k] = kwargs[k]
            else:
                raise ValueError(
                    "Missing key {} for signature {}".format(k, signature))

        return output

    @staticmethod
    def get_inputs(inputs):
        if len(inputs) == 3:
            x, a, e = inputs
            assert K.ndim(e) in (2, 3), "E must have rank 2 or 3"
        elif len(inputs) == 2:
            x, a = inputs
            e = None
        else:
            raise ValueError(
                "Expected 2 or 3 inputs tensors (X, A, E), got {}.".format(len(inputs)))
        assert K.ndim(x) in (2, 3), "X must have rank 2 or 3"
        assert K.is_sparse(a), "A must be a SparseTensor"
        assert K.ndim(a) == 2, "A must have rank 2"

        return x, a, e

    def get_config(self):
        mp_config = {"aggregate": serialize_scatter(self.agg)}
        keras_config = {}
        for key in self.kwargs_keys:
            keras_config[key] = serialize_kwarg(key, getattr(self, key))
        base_config = super().get_config()

        return {**base_config, **keras_config, **mp_config, **self.config}

    @property
    def config(self):
        return {}

    @staticmethod
    def preprocess(a):
        return a


class AGNNConv(MessagePassing):
    r"""
    An Attention-based Graph Neural Network (AGNN) from the paper

    > [Attention-based Graph Neural Network for Semi-supervised Learning](https://arxiv.org/abs/1803.03735)<br>
    > Kiran K. Thekumparampil et al.

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes:
    $$
        \X' = \P\X
    $$
    where
    $$
        \P_{ij} = \frac{
            \exp \left( \beta \cos \left( \x_i, \x_j \right) \right)
        }{
            \sum\limits_{k \in \mathcal{N}(i) \cup \{ i \}}
            \exp \left( \beta \cos \left( \x_i, \x_k \right) \right)
        }
    $$
    and \(\beta\) is a trainable parameter.

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input.

    **Arguments**

    - `trainable`: boolean, if True, then beta is a trainable parameter.
    Otherwise, beta is fixed to 1;
    - `activation`: activation function;
    """

    def __init__(self, trainable=True, aggregate="sum", activation=None, **kwargs):
        super().__init__(aggregate=aggregate, activation=activation, **kwargs)
        self.trainable = trainable

    def build(self, input_shape):
        assert len(input_shape) >= 2
        if self.trainable:
            self.beta = self.add_weight(
                shape=(1,), initializer="ones", name="beta")
        else:
            self.beta = K.constant(1.0)
        self.built = True

    def call(self, inputs, **kwargs):
        x, a, _ = self.get_inputs(inputs)
        x_norm = K.l2_normalize(x, axis=-1)
        output = self.propagate(x, a, x_norm=x_norm)
        output = self.activation(output)

        return output

    def message(self, x, x_norm=None):
        x_j = self.get_j(x)
        x_norm_i = self.get_i(x_norm)
        x_norm_j = self.get_j(x_norm)
        alpha = self.beta * tf.reduce_sum(x_norm_i * x_norm_j, axis=-1)

        if len(tf.shape(alpha)) == 2:
            alpha = tf.transpose(alpha)  # For mixed mode
        alpha = ops.unsorted_segment_softmax(alpha, self.index_i, self.n_nodes)
        if len(tf.shape(alpha)) == 2:
            alpha = tf.transpose(alpha)  # For mixed mode
        alpha = alpha[..., None]

        return alpha * x_j

    @property
    def config(self):
        return {
            "trainable": self.trainable,
        }


class CrystalConv(MessagePassing):
    r"""
    A crystal graph convolutional layer from the paper

    > [Crystal Graph Convolutional Neural Networks for an Accurate and
    Interpretable Prediction of Material Properties](https://arxiv.org/abs/1710.10324)<br>
    > Tian Xie and Jeffrey C. Grossman

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes:
    $$
        \x_i' = \x_i + \sum\limits_{j \in \mathcal{N}(i)} \sigma \left( \z_{ij}
        \W^{(f)} + \b^{(f)} \right) \odot \g \left( \z_{ij} \W^{(s)} + \b^{(s)}
        \right)
    $$
    where \(\z_{ij} = \X_i \| \X_j \| \E_{ij} \), \(\sigma\) is a sigmoid
    activation, and \(g\) is the activation function (defined by the `activation`
    argument).

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.
    - Edge features of shape `(num_edges, n_edge_features)`.

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `channels`.

    **Arguments**

    - `channels`: integer, number of output channels;
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        channels,
        aggregate="sum",
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels

    def build(self, input_shape):
        assert len(input_shape) >= 2
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
        )
        self.dense_f = Dense(
            self.channels, activation="sigmoid", **layer_kwargs)
        self.dense_s = Dense(
            self.channels, activation=self.activation, **layer_kwargs)

        self.built = True

    def message(self, x, e=None):
        x_i = self.get_i(x)
        x_j = self.get_j(x)

        to_concat = [x_i, x_j]
        if e is not None:
            to_concat += [e]
        z = K.concatenate(to_concat, axis=-1)
        output = self.dense_s(z) * self.dense_f(z)

        return output

    def update(self, embeddings, x=None):
        return x + embeddings

    @property
    def config(self):
        return {"channels": self.channels}


class EdgeConv(MessagePassing):
    r"""
    An edge convolutional layer from the paper

    > [Dynamic Graph CNN for Learning on Point Clouds](https://arxiv.org/abs/1801.07829)<br>
    > Yue Wang et al.

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes for each node \(i\):
    $$
        \x_i' = \sum\limits_{j \in \mathcal{N}(i)} \textrm{MLP}\big( \x_i \|
        \x_j - \x_i \big)
    $$
    where \(\textrm{MLP}\) is a multi-layer perceptron.

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `channels`.

    **Arguments**

    - `channels`: integer, number of output channels;
    - `mlp_hidden`: list of integers, number of hidden units for each hidden
    layer in the MLP (if None, the MLP has only the output layer);
    - `mlp_activation`: activation for the MLP layers;
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        channels,
        mlp_hidden=None,
        mlp_activation="relu",
        aggregate="sum",
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels
        self.mlp_hidden = mlp_hidden if mlp_hidden else []
        self.mlp_activation = activations.get(mlp_activation)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
        )

        self.mlp = Sequential(
            [
                Dense(channels, self.mlp_activation, **layer_kwargs)
                for channels in self.mlp_hidden
            ]
            + [
                Dense(
                    self.channels,
                    self.activation,
                    use_bias=self.use_bias,
                    **layer_kwargs
                )
            ]
        )

        self.built = True

    def message(self, x, **kwargs):
        x_i = self.get_i(x)
        x_j = self.get_j(x)
        return self.mlp(K.concatenate((x_i, x_j - x_i)))

    @property
    def config(self):
        return {
            "channels": self.channels,
            "mlp_hidden": self.mlp_hidden,
            "mlp_activation": self.mlp_activation,
        }


class GatedGraphConv(MessagePassing):
    r"""
    A gated graph convolutional layer from the paper

    > [Gated Graph Sequence Neural Networks](https://arxiv.org/abs/1511.05493)<br>
    > Yujia Li et al.

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes \(\x_i' = \h^{(L)}_i\) where:
    $$
    \begin{align}
        & \h^{(0)}_i = \x_i \| \mathbf{0} \\
        & \m^{(l)}_i = \sum\limits_{j \in \mathcal{N}(i)} \h^{(l - 1)}_j \W \\
        & \h^{(l)}_i = \textrm{GRU} \left(\m^{(l)}_i, \h^{(l - 1)}_i \right) \\
    \end{align}
    $$
    where \(\textrm{GRU}\) is a gated recurrent unit cell.

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`; note that
    `n_node_features` must be smaller or equal than `channels`.
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `channels`.

    **Arguments**

    - `channels`: integer, number of output channels;
    - `n_layers`: integer, number of iterations with the GRU cell;
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        channels,
        n_layers,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels
        self.n_layers = n_layers

    def build(self, input_shape):
        assert len(input_shape) >= 2
        self.kernel = self.add_weight(
            name="kernel",
            shape=(self.n_layers, self.channels, self.channels),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )
        self.rnn = GRUCell(
            self.channels,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            activity_regularizer=self.activity_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
            use_bias=self.use_bias,
        )
        self.built = True

    def call(self, inputs):
        x, a, _ = self.get_inputs(inputs)
        F = K.int_shape(x)[-1]

        to_pad = self.channels - F
        ndims = len(tf.shape(x)) - 1
        output = tf.pad(x, [[0, 0]] * ndims + [[0, to_pad]])
        for i in range(self.n_layers):
            m = tf.matmul(output, self.kernel[i])
            m = self.propagate(m, a)
            output = self.rnn(m, [output])[0]

        output = self.activation(output)
        return output

    @property
    def config(self):
        return {
            "channels": self.channels,
            "n_layers": self.n_layers,
        }


class GeneralConv(MessagePassing):
    r"""
    A general convolutional layer from the paper

    > [Design Space for Graph Neural Networks](https://arxiv.org/abs/2011.08843)<br>
    > Jiaxuan You et al.

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes:
    $$
        \x_i' = \mathrm{Agg} \left( \left\{ \mathrm{Act} \left( \mathrm{Dropout}
        \left( \mathrm{BN} \left( \x_j \W + \b \right) \right) \right),
        j \in \mathcal{N}(i) \right\} \right)
    $$

    where \( \mathrm{Agg} \) is an aggregation function for the messages,
    \( \mathrm{Act} \) is an activation function, \( \mathrm{Dropout} \)
    applies dropout to the node features, and \( \mathrm{BN} \) applies batch
    normalization to the node features.

    This layer supports the PReLU activation via the 'prelu' keyword.

    The default parameters of this layer are selected according to the best
    results obtained in the paper, and should provide a good performance on
    many node-level and graph-level tasks, without modifications.
    The defaults are as follows:

    - 256 channels
    - Batch normalization
    - No dropout
    - PReLU activation
    - Sum aggregation

    If you are uncertain about which layers to use for your GNN, this is a
    safe choice. Check out the original paper for more specific configurations.

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `channels`.

    **Arguments**

    - `channels`: integer, number of output channels;
    - `batch_norm`: bool, whether to use batch normalization;
    - `dropout`: float, dropout rate;
    - `aggregate`: string or callable, an aggregation function. Supported
    aggregations: 'sum', 'mean', 'max', 'min', 'prod'.
    - `activation`: activation function. This layer also supports the
    advanced activation PReLU by passing `activation='prelu'`.
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        channels=256,
        batch_norm=True,
        dropout=0.0,
        aggregate="sum",
        activation="prelu",
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            activation=None,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels
        self.dropout_rate = dropout
        self.use_batch_norm = batch_norm
        if activation == "prelu" or "prelu" in kwargs:
            self.activation = PReLU()
        else:
            self.activation = activations.get(activation)

    def build(self, input_shape):
        input_dim = input_shape[0][-1]
        self.dropout = Dropout(self.dropout_rate)
        if self.use_batch_norm:
            self.batch_norm = BatchNormalization()
        self.kernel = self.add_weight(
            shape=(input_dim, self.channels),
            initializer=self.kernel_initializer,
            name="kernel",
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )
        if self.use_bias:
            self.bias = self.add_weight(
                shape=(self.channels,),
                initializer=self.bias_initializer,
                name="bias",
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        self.built = True

    def call(self, inputs, **kwargs):
        x, a, _ = self.get_inputs(inputs)

        # TODO: a = add_self_loops(a)

        x = tf.matmul(x, self.kernel)
        if self.use_bias:
            x = tf.nn.bias_add(x, self.bias)
        if self.use_batch_norm:
            x = self.batch_norm(x)
        x = self.dropout(x)
        x = self.activation(x)

        return self.propagate(x, a)

    @property
    def config(self):
        config = {
            "channels": self.channels,
        }
        if self.activation.__class__.__name__ == "PReLU":
            config["prelu"] = True

        return config


class GINConv(MessagePassing):
    r"""
    A Graph Isomorphism Network (GIN) from the paper

    > [How Powerful are Graph Neural Networks?](https://arxiv.org/abs/1810.00826)<br>
    > Keyulu Xu et al.

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes for each node \(i\):
    $$
        \x_i' = \textrm{MLP}\big( (1 + \epsilon) \cdot \x_i + \sum\limits_{j
        \in \mathcal{N}(i)} \x_j \big)
    $$
    where \(\textrm{MLP}\) is a multi-layer perceptron.

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `channels`.

    **Arguments**

    - `channels`: integer, number of output channels;
    - `epsilon`: unnamed parameter, see the original paper and the equation
    above.
    By setting `epsilon=None`, the parameter will be learned (default behaviour).
    If given as a value, the parameter will stay fixed.
    - `mlp_hidden`: list of integers, number of hidden units for each hidden
    layer in the MLP (if None, the MLP has only the output layer);
    - `mlp_activation`: activation for the MLP layers;
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        channels,
        epsilon=None,
        mlp_hidden=None,
        mlp_activation="relu",
        aggregate="sum",
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels
        self.epsilon = epsilon
        self.mlp_hidden = mlp_hidden if mlp_hidden else []
        self.mlp_activation = activations.get(mlp_activation)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
        )

        self.mlp = Sequential(
            [
                Dense(channels, self.mlp_activation, **layer_kwargs)
                for channels in self.mlp_hidden
            ]
            + [
                Dense(
                    self.channels,
                    self.activation,
                    use_bias=self.use_bias,
                    **layer_kwargs
                )
            ]
        )

        if self.epsilon is None:
            self.eps = self.add_weight(
                shape=(1,), initializer="zeros", name="eps")
        else:
            # If epsilon is given, keep it constant
            self.eps = K.constant(self.epsilon)

        self.built = True

    def call(self, inputs, **kwargs):
        x, a, _ = self.get_inputs(inputs)
        output = self.mlp((1.0 + self.eps) * x + self.propagate(x, a))

        return output

    @property
    def config(self):
        return {
            "channels": self.channels,
            "epsilon": self.epsilon,
            "mlp_hidden": self.mlp_hidden,
            "mlp_activation": self.mlp_activation,
        }


class GraphSageConv(MessagePassing):
    r"""
    A GraphSAGE layer from the paper

    > [Inductive Representation Learning on Large Graphs](https://arxiv.org/abs/1706.02216)<br>
    > William L. Hamilton et al.

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes:
    $$
        \X' = \big[ \textrm{AGGREGATE}(\X) \| \X \big] \W + \b; \\
        \X' = \frac{\X'}{\|\X'\|}
    $$
    where \( \textrm{AGGREGATE} \) is a function to aggregate a node's
    neighbourhood. The supported aggregation methods are: sum, mean,
    max, min, and product.

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape as the input, but with the last
    dimension changed to `channels`.

    **Arguments**

    - `channels`: number of output channels;
    - `aggregate_op`: str, aggregation method to use (`'sum'`, `'mean'`,
    `'max'`, `'min'`, `'prod'`);
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.

    """

    def __init__(
        self,
        channels,
        aggregate="mean",
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels

    def build(self, input_shape):
        assert len(input_shape) >= 2
        input_dim = input_shape[0][-1]
        self.kernel = self.add_weight(
            shape=(2 * input_dim, self.channels),
            initializer=self.kernel_initializer,
            name="kernel",
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )
        if self.use_bias:
            self.bias = self.add_weight(
                shape=(self.channels,),
                initializer=self.bias_initializer,
                name="bias",
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        else:
            self.bias = None
        self.built = True

    def call(self, inputs):
        x, a, _ = self.get_inputs(inputs)
        a = ops.add_self_loops(a)

        aggregated = self.propagate(x, a)
        output = K.concatenate([x, aggregated])
        output = K.dot(output, self.kernel)

        if self.use_bias:
            output = K.bias_add(output, self.bias)
        output = K.l2_normalize(output, axis=-1)
        if self.activation is not None:
            output = self.activation(output)

        return output

    @property
    def config(self):
        return {"channels": self.channels}


class TAGConv(MessagePassing):
    r"""
    A Topology Adaptive Graph Convolutional layer (TAG) from the paper

    > [Topology Adaptive Graph Convolutional Networks](https://arxiv.org/abs/1710.10370)<br>
    > Jian Du et al.

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes:
    $$
        \Z = \sum\limits_{k=0}^{K} \D^{-1/2}\A^k\D^{-1/2}\X\W^{(k)}
    $$

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `channels`.

    **Arguments**

    - `channels`: integer, number of output channels;
    - `K`: the order of the layer (i.e., the layer will consider a K-hop
    neighbourhood for each node);
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        channels,
        K=3,
        aggregate="sum",
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels
        self.K = K
        self.linear = Dense(
            channels,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
        )

    def build(self, input_shape):
        assert len(input_shape) >= 2
        self.built = True

    def call(self, inputs, **kwargs):
        x, a, _ = self.get_inputs(inputs)
        edge_weight = a.values

        output = [x]
        for k in range(self.K):
            output.append(self.propagate(x, a, edge_weight=edge_weight))
        output = K.concatenate(output)

        return self.linear(output)

    def message(self, x, edge_weight=None):
        x_j = self.get_j(x)
        return edge_weight[:, None] * x_j

    @property
    def config(self):
        return {
            "channels": self.channels,
        }

    @staticmethod
    def preprocess(a):
        return normalized_adjacency(a)



import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K
from tensorflow.keras import initializers, layers
from tensorflow.keras.layers import Dense, Dropout, Layer
from tensorflow.keras.regularizers import l2


class SelfMean(Layer):
    """
    其功能是对2d或3d的tensor，指定一个axis进行求均值。例如[100,5,6]的矩阵，指定axis=1求均值，会变成[100,6]大小的矩阵。
    """

    def __init__(self, axis, **kwargs):
        self.supports_masking = True
        self.axis = axis
        super(SelfMean, self).__init__(**kwargs)

    def compute_mask(self, inputs, mask=None):
        return None

    def call(self, x, mask=None):
        if mask is not None:
            if K.ndim(x) != K.ndim(mask):
                mask = K.repeat(mask, x.shape[-1])
                mask = tf.transpose(mask, [0, 2, 1])
            mask = K.cast(mask, K.floatx())
            x = x * mask
            return K.sum(x, axis=self.axis) / K.sum(mask, axis=self.axis)

        return K.mean(x, axis=self.axis)

    def compute_output_shape(self, input_shape):
        output_shape = []
        for i in range(len(input_shape)):
            if i != self.axis:
                output_shape.append(input_shape[i])
        return tuple(output_shape)

    def get_config(self):
        config = {"axis": self.axis}
        config.update(super(SelfMean, self).get_config())
        return config


class SelfSum(Layer):
    def __init__(self, axis, **kwargs):
        self.supports_masking = True
        self.axis = axis
        super(SelfSum, self).__init__(**kwargs)

    def compute_mask(self, inputs, mask=None):
        return None

    def call(self, x, mask=None, *args, **kwargs):
        if mask is not None:
            # mask (batch, time)
            mask = K.cast(mask, K.floatx())
            if K.ndim(x) != K.ndim(mask):
                mask = K.repeat(mask, x.shape[-1])
                mask = tf.transpose(mask, [0, 2, 1])
            x = x * mask
            if K.ndim(x) == 2:
                x = K.expand_dims(x)
            return K.sum(x, axis=self.axis)
        else:
            if K.ndim(x) == 2:
                x = K.expand_dims(x)
            return K.sum(x, axis=self.axis)

    def compute_output_shape(self, input_shape):
        output_shape = []
        for i in range(len(input_shape)):
            if i != self.axis:
                output_shape.append(input_shape[i])
        if len(output_shape) == 1:
            output_shape.append(1)
        return tuple(output_shape)

    def get_config(self):
        config = {"axis": self.axis}
        config.update(super(SelfSum, self).get_config())
        return config


class MaskFlatten(Layer):
    def __init__(self, *args, **kwargs):
        self.supports_masking = True
        super(MaskFlatten, self).__init__(*args, **kwargs)

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return mask
        return K.batch_flatten(mask)

    def call(self, inputs, mask=None):
        return K.batch_flatten(inputs)

    def compute_output_shape(self, input_shape):
        return input_shape[0], np.prod(input_shape[1:])


class AutoEncoder(Layer):
    """
    https://keras.io/layers/writing-your-own-keras-layers/
    """

    def __init__(self,
                 hidden_layer=None,
                 name='Cross',
                 is_layer=False,
                 hidden_activation='relu',
                 output_activation='tanh',
                 **kwargs):
        super(AutoEncoder, self).__init__(name=name, **kwargs)
        if hidden_layer is None:
            hidden_layer = [128, 64, 32]

        self.is_layer = is_layer
        self.hidden_layer = hidden_layer
        self.hidden_activation = hidden_activation
        self.output_activation = output_activation
        self.encode_layer = self.decode_layer = None

    def build(self, input_shape):

        self.encode_layer = []
        self.decode_layer = []

        index = 0
        index_max = len(self.hidden_layer)
        for layer_num in self.hidden_layer[:-1]:
            index += 1
            self.encode_layer.append(Dense(layer_num,
                                           name='{}-encode-{}'.format(
                                               self.name, index),
                                           activation=self.hidden_activation
                                           ))
            self.decode_layer.insert(0, Dense(layer_num,
                                              name='{}-decode-{}'.format(
                                                  self.name, index_max - index),
                                              activation=self.hidden_activation
                                              ))
        self.encode_layer.append(Dense(self.hidden_layer[-1],
                                       name='{}-encode-{}'.format(
                                           self.name, index_max),
                                       activation=self.hidden_activation
                                       ))
        self.decode_layer.append(Dense(input_shape[-1],
                                       name='{}-decode-{}'.format(
                                           self.name, index_max),
                                       activation=self.output_activation
                                       ))

        super(AutoEncoder, self).build(input_shape)

    def call(self, inputs, **kwargs):
        layer = inputs
        for encoder in self.encode_layer:
            layer = encoder(layer)

        for decoder in self.decode_layer:
            layer = decoder(layer)

        return layer

    def __call__(self, inputs, **kwargs):
        if self.is_layer:
            return super(AutoEncoder, self).__call__(inputs, **kwargs)
        else:
            self.build(inputs.shape)
            return self.call(inputs)

    def compute_output_shape(self, input_shape):
        return input_shape


class CrossLayer(Layer):
    def __init__(self, layer_num=5, reg_weight=1e-4, reg_bias=1e-4, *args, **kwargs):
        super(CrossLayer, self).__init__(*args, **kwargs)
        self.layer_num = layer_num
        self.reg_weight = reg_weight
        self.reg_bias = reg_bias
        self.kernels = self.biases = None

    def build(self, input_shape):
        dim = int(input_shape[1])
        self.kernels = [self.add_weight(name='weight_' + str(i),
                                        shape=(dim, 1),
                                        initializer='random_uniform',
                                        regularizer=l2(self.reg_weight),
                                        trainable=True, ) for i in range(self.layer_num)]

        self.biases = [self.add_weight(name='bias_' + str(i),
                                       shape=(dim, 1),
                                       initializer='random_uniform',
                                       regularizer=l2(self.reg_bias),
                                       trainable=True, ) for i in range(self.layer_num)]

    def call(self, inputs, **kwargs):
        # (None, dim, 1)
        x_0 = tf.expand_dims(inputs, axis=2)
        # (None, dim, 1)
        x_l = x_0
        for i in range(self.layer_num):
            # (None, dim, dim)
            x_l1 = tf.tensordot(x_l, self.kernels[i], axes=[1, 0])
            # (None, dim, 1)
            x_l = tf.matmul(x_0, x_l1) + self.biases[i] + x_l

        # (None, dim)
        x_l = tf.squeeze(x_l, axis=2)
        return x_l


class Linear(Dense):
    """
    Linear Part
    """

    def __init__(self, *args, **kwargs):
        super(Linear, self).__init__(1, activation=None, *args, **kwargs)


class DNN(Layer):
    """
        Deep Neural Network
        """

    def __init__(self,
                 hidden_units,
                 activation='relu',
                 dropout=0.,
                 name='DNN',
                 *args,
                 **kwargs):
        """
        :param hidden_units: A list. Neural network hidden units.
        :param activation: A string. Activation function of dnn.
        :param dropout: A scalar. Dropout number.
        """
        super(DNN, self).__init__(*args, **kwargs)
        self.hidden_units = hidden_units
        self.activation = activation
        self.dropout = dropout
        self.dnn_network = None
        self.args = args
        self.kwargs = kwargs

    def build(self, input_shape):
        self.dnn_network = [Dense(units=unit, activation=self.activation, *self.args, **self.kwargs)
                            for unit in self.hidden_units]
        self.dropout = Dropout(self.dropout)
        super(DNN, self).build(input_shape)

    def call(self, inputs, **kwargs):
        x = inputs
        for dnn in self.dnn_network:
            x = dnn(x)
        x = self.dropout(x)
        return x



import demjson
import numpy as np
import pandas as pd
from pandas import DataFrame
from sklearn.preprocessing import (Binarizer, LabelBinarizer, LabelEncoder,
                                   MaxAbsScaler, MinMaxScaler, Normalizer,
                                   OneHotEncoder, RobustScaler, StandardScaler)


class FeaturePreProcessing:
    def __init__(self):
        self.feature_dict = {}

    def _df_transform(self, dataframe: DataFrame, field, name):
        if name in ['Normalizer']:
            return 2, [dataframe[field].values]
        elif name in ['a']:
            return 1, np.transpose(dataframe[field].values)
        else:
            return 1, np.transpose([dataframe[field].values])

    def _field_converse(self, dataframe: DataFrame, fields=None) -> dict:
        if isinstance(fields, str) or isinstance(fields, list):
            new_fields = {}
            for col in dataframe.columns:
                new_fields[col] = fields
            return self._field_converse(dataframe, new_fields)
        elif isinstance(fields, dict):
            for field_k, field_v in fields.items():
                if isinstance(field_v, str):
                    fields[field_k] = {
                        field_v: {}
                    }
                elif isinstance(field_v, list):
                    fields[field_k] = {}
                    for field in field_v:
                        fields[field_k][field] = {}
            return fields
        else:
            print("error")
            return fields

    def _fit(self, dataframe: DataFrame, field_k, field_v=None, *args, **kwargs):
        if field_v in ('stand', 'StandardScaler', None):
            value = StandardScaler(*args, **kwargs)
        elif field_v in ('minmax', 'MinMaxScaler'):
            value = MinMaxScaler(*args, **kwargs)
        elif field_v in ('maxabs', 'MaxAbsScaler'):
            value = MaxAbsScaler(*args, **kwargs)
        elif field_v in ('norm', 'Normalizer'):
            value = Normalizer(*args, **kwargs)
        elif field_v in ('encoder', 'LabelEncoder'):
            value = LabelEncoder()
        elif field_v in ('robust', 'RobustScaler'):
            value = RobustScaler(*args, **kwargs)
        elif field_v in ('Binarizer', 'Binarizer'):
            value = Binarizer(*args, **kwargs)
        elif field_v in ('LabelBinarizer', 'LabelBinarizer'):
            value = LabelBinarizer(*args, **kwargs)
        elif field_v in ('OneHotEncoder', 'OneHotEncoder'):
            value = OneHotEncoder(*args, **kwargs)
        else:
            value = StandardScaler(*args, **kwargs)

        _, df = self._df_transform(dataframe, field_k, field_v)
        value.fit(df, *args, **kwargs)
        if field_k in self.feature_dict.keys():
            self.feature_dict[field_k][field_v] = value
        else:
            self.feature_dict[field_k] = {}
            self.feature_dict[field_k][field_v] = value

    def _transform(self, dataframe: DataFrame, field_k, field_v=None, *args, **kwargs):
        num, df = self._df_transform(dataframe, field_k, field_v)
        if num == 2:
            dataframe[field_k] = self.feature_dict[field_k][field_v].transform(df)[
                0]
        else:
            dataframe[field_k] = self.feature_dict[field_k][field_v].transform(
                df)

    def fit(self, dataframe: DataFrame, fields=None, *args, **kwargs):
        fields = self._field_converse(dataframe, fields)
        for field, params in fields.items():
            for process, v in params.items():
                self._fit(dataframe, field, process, **v)

    def transform(self, dataframe: DataFrame, fields=None, *args, **kwargs):
        fields = self._field_converse(dataframe, fields)
        for field, params in fields.items():
            for process, v in params.items():
                self._transform(dataframe, field, process, **v)
        return dataframe

    def fit_transform(self, dataframe: DataFrame, fields=None, *args, **kwargs):
        fields = self._field_converse(dataframe, fields)
        for field, params in fields.items():
            for process, v in params.items():
                self._fit(dataframe, field, process, **v)
                self._transform(dataframe, field, process, **v)
        return dataframe

    def inverse_transform(self, *args, **kwargs):
        pass


class FeatureDictManage:
    """
    主要作用：将DataFrame中的离散字段映射成递增的数字，
    主要包含两步：
        1. 构建映射关系：构建离散字段->ID的映射
        2. 字段映射：将字段映射到相应的ID
    """

    def __init__(self):
        self.feature_size = {}
        self.feature_map = {}

    @staticmethod
    def _field_converse(dataframe: DataFrame, fields=None) -> dict:
        if fields is None:
            fields = dict(zip(dataframe.columns, dataframe.columns))
        elif isinstance(fields, str):
            fields = {fields: fields}
        elif isinstance(fields, list):
            fields = dict(zip(fields, fields))
        return fields

    def fit(self, dataframe: DataFrame, fields=None):
        """
        构建映射关系
        :param dataframe: 数据
        :param fields: 需要构建映射关系的字段，可以是str,list,map  {data_field : save_field}
        """
        fields = self._field_converse(dataframe, fields)

        for field_k, field_v in fields.items():
            if field_v not in self.feature_size.keys():
                self.feature_map[field_v] = {'': 0}
                self.feature_size[field_v] = 1

            field_list = set(np.array(dataframe[field_k].values.tolist()).reshape(1, -1)[0]) - set(
                self.feature_map.get(field_v, []))
            field_list = list(field_list)
            field_list.sort()
            size = self.feature_size.get(field_v, 0)
            d = dict(
                zip(field_list, [i + size for i in range(len(field_list))]))

            self.feature_map[field_v].update(d)
            self.feature_size[field_v] = len(self.feature_map[field_v])

    def transform(self, dataframe: DataFrame, fields: dict = None) -> DataFrame:
        """
        字段映射
        :param dataframe:
        :param fields:
        :return:
        """
        fields = self._field_converse(dataframe, fields)

        for field_k, field_v in fields.items():
            if field_v not in self.feature_map.keys():
                continue

            if isinstance(dataframe[field_k].values[0], list):
                dataframe[field_k] = dataframe[field_k].apply(
                    lambda x: [self.feature_map[field_v].get(i, 0) for i in x])
            else:
                dataframe[field_k] = dataframe[field_k].apply(
                    lambda x: self.feature_map[field_v][x])
        return dataframe

    def fit_transform(self, dataframe: DataFrame, fields: dict = None) -> DataFrame:
        self.fit(dataframe, fields)
        return self.transform(dataframe, fields)



import tensorflow as tf
from tensorflow.keras.layers import MaxPool2D,Conv2D

import notekeras.models.yolo4.core.common as common
from notekeras.component.image.core import SPPNet
from notekeras.component.yolo.core import YoloConv, YoloResBlock, YoloCSPX


def darknet53(input_data):
    input_data = YoloConv(filters=32, kernel_size=3, down_sample=True)(input_data)

    input_data = YoloConv(filters=64, kernel_size=3, down_sample=True)(input_data)
    for i in range(1):
        input_data = YoloResBlock(input_channel=64, filter_num1=32, filter_num2=64)(input_data)

    input_data = YoloConv(filters=128, kernel_size=3, down_sample=True)(input_data)
    for i in range(2):
        input_data = YoloResBlock(input_channel=128, filter_num1=64, filter_num2=128)(input_data)

    input_data = YoloConv(filters=256, kernel_size=3, down_sample=True)(input_data)
    for i in range(8):
        input_data = YoloResBlock(input_channel=256, filter_num1=128, filter_num2=256)(input_data)

    route_1 = input_data

    input_data = YoloConv(filters=512, kernel_size=3, down_sample=True)(input_data)
    for i in range(8):
        input_data = YoloResBlock(input_channel=512, filter_num1=256, filter_num2=512)(input_data)

    route_2 = input_data

    input_data = YoloConv(filters=1024, kernel_size=3, down_sample=True)(input_data)
    for i in range(4):
        input_data = YoloResBlock(input_channel=1024, filter_num1=512, filter_num2=1024)(input_data)

    return route_1, route_2, input_data


def cspdarknet53(input_data):
    input_data = YoloConv(filters=32, kernel_size=3, activate_type='mish')(input_data)
    input_data = YoloConv(filters=64, kernel_size=3, activate_type='mish', down_sample=True)(input_data)
    route_data = YoloConv(filters=64, kernel_size=1, activate_type='mish')(input_data)
    input_data = YoloConv(filters=64, kernel_size=1, activate_type='mish')(input_data)

    for i in range(1):
        input_data = YoloResBlock(input_channel=64, filter_num1=32, filter_num2=64, activate_type='mish')(input_data)
    input_data = YoloConv(filters=64, kernel_size=1, activate_type='mish')(input_data)
    input_data = tf.concat([input_data, route_data], axis=-1)

    #################
    input_data = YoloConv(filters=64, kernel_size=1, activate_type='mish')(input_data)
    input_data = YoloConv(filters=128, kernel_size=3, activate_type='mish', down_sample=True)(input_data)
    input_data = YoloCSPX(input_data, rec_num=2, filters=64, kernel_size=1, activate_type='mish')

    #################
    input_data = YoloConv(filters=128, kernel_size=1, activate_type='mish')(input_data)
    input_data = YoloConv(filters=256, kernel_size=3, activate_type='mish', down_sample=True)(input_data)
    input_data = YoloCSPX(input_data, rec_num=8, filters=128, kernel_size=1, activate_type='mish')

    #################
    input_data = YoloConv(filters=256, kernel_size=1, activate_type='mish')(input_data)
    route_1 = input_data
    input_data = YoloConv(filters=512, kernel_size=3, activate_type='mish', down_sample=True)(input_data)
    input_data = YoloCSPX(input_data, rec_num=8, filters=256, kernel_size=1, activate_type='mish')

    #################
    input_data = YoloConv(filters=512, kernel_size=1, activate_type='mish')(input_data)
    route_2 = input_data
    input_data = YoloConv(filters=1024, kernel_size=3, activate_type='mish', down_sample=True)(input_data)
    input_data = YoloCSPX(input_data, rec_num=4, filters=512, kernel_size=1, activate_type='mish')

    #################
    input_data = YoloConv(filters=1024, kernel_size=1, activate_type='mish')(input_data)

    input_data = YoloConv(filters=512, kernel_size=1)(input_data)
    input_data = YoloConv(filters=1024, kernel_size=3)(input_data)
    input_data = YoloConv(filters=512, kernel_size=1)(input_data)

    input_data = SPPNet(pool_size=[13, 9, 5, 1], padding='SAME', strides=1)(input_data)

    input_data = YoloConv(filters=512, kernel_size=1)(input_data)
    input_data = YoloConv(filters=1024, kernel_size=3)(input_data)
    route_3 = YoloConv(filters=512, kernel_size=1)(input_data)

    return route_1, route_2, route_3


def cspdarknet53_tiny(input_data):
    input_data = YoloConv(filters=32, kernel_size=3, down_sample=True)(input_data)
    input_data = YoloConv(filters=64, kernel_size=3, down_sample=True)(input_data)
    input_data = YoloConv(filters=64, kernel_size=3)(input_data)

    route = input_data
    input_data = common.route_group(input_data, 2, 1)
    input_data = YoloConv(filters=32, kernel_size=3)(input_data)
    route_1 = input_data
    input_data = YoloConv(filters=32, kernel_size=3)(input_data)
    input_data = tf.concat([input_data, route_1], axis=-1)
    input_data = YoloConv(filters=64, kernel_size=1)(input_data)
    input_data = tf.concat([route, input_data], axis=-1)
    input_data = MaxPool2D(2, 2, 'same')(input_data)
    input_data = YoloConv(filters=128, kernel_size=3)(input_data)

    route = input_data
    input_data = common.route_group(input_data, 2, 1)
    input_data = YoloConv(filters=64, kernel_size=3)(input_data)
    route_1 = input_data
    input_data = YoloConv(filters=64, kernel_size=3)(input_data)
    input_data = tf.concat([input_data, route_1], axis=-1)
    input_data = YoloConv(filters=128, kernel_size=1)(input_data)
    input_data = tf.concat([route, input_data], axis=-1)
    input_data = MaxPool2D(2, 2, 'same')(input_data)
    input_data = YoloConv(filters=256, kernel_size=3)(input_data)

    route = input_data
    input_data = common.route_group(input_data, 2, 1)
    input_data = YoloConv(filters=128, kernel_size=3)(input_data)
    route_1 = input_data
    input_data = YoloConv(filters=128, kernel_size=3)(input_data)
    input_data = tf.concat([input_data, route_1], axis=-1)
    input_data = YoloConv(filters=256, kernel_size=1)(input_data)
    route_1 = input_data
    input_data = tf.concat([route, input_data], axis=-1)
    input_data = MaxPool2D(2, 2, 'same')(input_data)
    input_data = YoloConv(filters=512, kernel_size=3)(input_data)

    return route_1, input_data


def darknet53_tiny(input_data):
    input_data = YoloConv(filters=16, kernel_size=3)(input_data)
    input_data = MaxPool2D(2, 2, 'same')(input_data)
    input_data = YoloConv(filters=32, kernel_size=3)(input_data)
    input_data = MaxPool2D(2, 2, 'same')(input_data)
    input_data = YoloConv(filters=64, kernel_size=3)(input_data)
    input_data = MaxPool2D(2, 2, 'same')(input_data)
    input_data = YoloConv(filters=128, kernel_size=3)(input_data)
    input_data = MaxPool2D(2, 2, 'same')(input_data)
    input_data = YoloConv(filters=256, kernel_size=3)(input_data)

    route_1 = input_data
    input_data = MaxPool2D(2, 2, 'same')(input_data)
    input_data = YoloConv(filters=512, kernel_size=3)(input_data)
    input_data = MaxPool2D(2, 1, 'same')(input_data)
    input_data = YoloConv(filters=1024, kernel_size=3)(input_data)

    return route_1, input_data



import os

from tensorflow import keras
from tensorflow.keras.backend import *

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

__all__ = [
    'keras', 'utils', 'activations', 'applications', 'backend', 'datasets',
    'layers', 'preprocessing', 'wrappers', 'callbacks', 'constraints', 'initializers',
    'metrics', 'models', 'losses', 'optimizers', 'regularizers',
    'Dense', 'plot_model',
    'Layer', 'Model',
]

# keras.backend.set_floatx('float64')
utils = keras.utils
activations = keras.activations
applications = keras.applications
backend = keras.backend
datasets = keras.datasets

preprocessing = keras.preprocessing
wrappers = keras.wrappers
callbacks = keras.callbacks
constraints = keras.constraints
initializers = keras.initializers
metrics = keras.metrics
models = keras.models
losses = keras.losses
optimizers = keras.optimizers
regularizers = keras.regularizers
plot_model = keras.utils.plot_model

layers = keras.layers
Layer = layers.Layer
Dense = layers.Dense
Model = keras.models.Model



import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow import keras
from tensorflow.keras.layers import Layer, Embedding, Add


class TokenEmbedding(Embedding):
    """Embedding layer with weights returned."""

    def compute_output_shape(self, input_shape):
        return [super(TokenEmbedding, self).compute_output_shape(input_shape), (self.input_dim, self.output_dim)]

    def compute_mask(self, inputs, mask=None):
        return [super(TokenEmbedding, self).compute_mask(inputs, mask), None]

    def call(self, inputs):
        return [super(TokenEmbedding, self).call(inputs), tf.identity(self.embeddings)]


class EmbeddingSimilarity(Layer):
    """Calculate similarity between features and token embeddings with bias term."""

    def __init__(self,
                 initializer='zeros',
                 regularizer=None,
                 constraint=None,
                 **kwargs):
        """Initialize the layer.

        :param output_dim: Same as embedding output dimension.
        :param initializer: Initializer for bias.
        :param regularizer: Regularizer for bias.
        :param constraint: Constraint for bias.
        :param kwargs: Arguments for parent class.
        """
        super(EmbeddingSimilarity, self).__init__(**kwargs)
        self.supports_masking = True
        self.initializer = keras.initializers.get(initializer)
        self.regularizer = keras.regularizers.get(regularizer)
        self.constraint = keras.constraints.get(constraint)
        self.bias = None

    def get_config(self):
        config = {
            'initializer': keras.initializers.serialize(self.initializer),
            'regularizer': keras.regularizers.serialize(self.regularizer),
            'constraint': keras.constraints.serialize(self.constraint),
        }
        base_config = super(EmbeddingSimilarity, self).get_config()
        base_config.update(config)
        return base_config

    def build(self, input_shape):
        self.bias = self.add_weight(
            shape=(int(input_shape[1][0]),),
            initializer=self.initializer,
            regularizer=self.regularizer,
            constraint=self.constraint,
            name='bias',
        )
        super(EmbeddingSimilarity, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        return input_shape[0][:2] + (input_shape[1][0],)

    def compute_mask(self, inputs, mask=None):
        return mask[0]

    def call(self, inputs, mask=None, **kwargs):
        inputs, embeddings = inputs
        outputs = K.bias_add(K.dot(inputs, K.transpose(embeddings)), self.bias)
        return keras.activations.softmax(outputs)


class EmbeddingSim(Layer):
    """Calculate similarity between features and token embeddings with bias term."""

    def __init__(self,
                 use_bias=True,
                 initializer='zeros',
                 regularizer=None,
                 constraint=None,
                 stop_gradient=False,
                 **kwargs):
        """Initialize the layer.

        :param output_dim: Same as embedding output dimension.
        :param use_bias: Whether to use bias term.
        :param initializer: Initializer for bias.
        :param regularizer: Regularizer for bias.
        :param constraint: Constraint for bias.
        :param stop_gradient: Whether to stop gradient for input embedding.
        :param kwargs: Arguments for parent class.
        """
        super(EmbeddingSim, self).__init__(**kwargs)
        self.supports_masking = True
        self.use_bias = use_bias
        self.initializer = keras.initializers.get(initializer)
        self.regularizer = keras.regularizers.get(regularizer)
        self.constraint = keras.constraints.get(constraint)
        self.stop_gradient = stop_gradient
        self.bias = None

    def get_config(self):
        config = {
            'use_bias': self.use_bias,
            'initializer': keras.initializers.serialize(self.initializer),
            'regularizer': keras.regularizers.serialize(self.regularizer),
            'constraint': keras.constraints.serialize(self.constraint),
            'stop_gradient': self.stop_gradient,
        }
        base_config = super(EmbeddingSim, self).get_config()
        base_config.update(config)
        return base_config

    def build(self, input_shape):
        if self.use_bias:
            embed_shape = input_shape[1]
            token_num = int(embed_shape[0])
            self.bias = self.add_weight(shape=(token_num,),
                                        initializer=self.initializer,
                                        regularizer=self.regularizer,
                                        constraint=self.constraint,
                                        name='bias',
                                        )
        super(EmbeddingSim, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        feature_shape, embed_shape = input_shape
        token_num = embed_shape[0]
        return feature_shape[:-1] + (token_num,)

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return None
        return mask[0]

    def call(self, inputs, mask=None, **kwargs):
        inputs, embeddings = inputs
        if self.stop_gradient:
            embeddings = K.stop_gradient(embeddings)
        outputs = K.dot(inputs, K.transpose(embeddings))
        if self.use_bias:
            outputs = K.bias_add(outputs, self.bias)
        return keras.activations.softmax(outputs)


class EmbeddingRet(Embedding):
    """
    Embedding layer with weights returned.
    """

    def compute_output_shape(self, input_shape):
        return [super(EmbeddingRet, self).compute_output_shape(input_shape), (self.input_dim, self.output_dim), ]

    def compute_mask(self, inputs, mask=None):
        return [super(EmbeddingRet, self).compute_mask(inputs, mask), None, ]

    def call(self, inputs):
        return [super(EmbeddingRet, self).call(inputs), tf.identity(self.embeddings), ]


class PositionEmbedding(Layer):
    """Turn integers (positions) into dense vectors of fixed size.
    eg. [[-4], [10]] -> [[0.25, 0.1], [0.6, -0.2]]

    Expand mode: negative integers (relative position) could be used in this mode.
        # Input shape
            2D tensor with shape: `(batch_size, sequence_length)`.

        # Output shape
            3D tensor with shape: `(batch_size, sequence_length, output_dim)`.

    Add mode:
        # Input shape
            3D tensor with shape: `(batch_size, sequence_length, feature_dim)`.

        # Output shape
            3D tensor with shape: `(batch_size, sequence_length, feature_dim)`.

    Concat mode:
        # Input shape
            3D tensor with shape: `(batch_size, sequence_length, feature_dim)`.

        # Output shape
            3D tensor with shape: `(batch_size, sequence_length, feature_dim + output_dim)`.
    """
    MODE_EXPAND = 'expand'
    MODE_ADD = 'add'
    MODE_CONCAT = 'concat'

    def __init__(self,
                 input_dim,
                 output_dim,
                 mode=MODE_EXPAND,
                 embeddings_initializer='uniform',
                 embeddings_regularizer=None,
                 activity_regularizer=None,
                 embeddings_constraint=None,
                 mask_zero=False,
                 **kwargs):
        """
        :param input_dim: The maximum absolute value of positions.
        :param output_dim: The embedding dimension.
        :param embeddings_initializer:
        :param embeddings_regularizer:
        :param activity_regularizer:
        :param embeddings_constraint:
        :param mask_zero: The index that represents padding. Only works in `append` mode.
        :param kwargs:
        """
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.mode = mode
        self.embeddings_initializer = keras.initializers.get(embeddings_initializer)
        self.embeddings_regularizer = keras.regularizers.get(embeddings_regularizer)
        self.activity_regularizer = keras.regularizers.get(activity_regularizer)
        self.embeddings_constraint = keras.constraints.get(embeddings_constraint)
        self.mask_zero = mask_zero
        self.supports_masking = mask_zero is not False

        self.embeddings = None
        super(PositionEmbedding, self).__init__(**kwargs)

    def get_config(self):
        config = {'input_dim': self.input_dim,
                  'output_dim': self.output_dim,
                  'mode': self.mode,
                  'embeddings_initializer': keras.initializers.serialize(self.embeddings_initializer),
                  'embeddings_regularizer': keras.regularizers.serialize(self.embeddings_regularizer),
                  'activity_regularizer': keras.regularizers.serialize(self.activity_regularizer),
                  'embeddings_constraint': keras.constraints.serialize(self.embeddings_constraint),
                  'mask_zero': self.mask_zero}
        base_config = super(PositionEmbedding, self).get_config()
        base_config.update(config)
        return base_config

    def build(self, input_shape):
        if self.mode == self.MODE_EXPAND:
            self.embeddings = self.add_weight(
                shape=(self.input_dim * 2 + 1, self.output_dim),
                initializer=self.embeddings_initializer,
                name='embeddings',
                regularizer=self.embeddings_regularizer,
                constraint=self.embeddings_constraint,
            )
        else:
            self.embeddings = self.add_weight(
                shape=(self.input_dim, self.output_dim),
                initializer=self.embeddings_initializer,
                name='embeddings',
                regularizer=self.embeddings_regularizer,
                constraint=self.embeddings_constraint,
            )
        super(PositionEmbedding, self).build(input_shape)

    def compute_mask(self, inputs, mask=None):
        if self.mode == self.MODE_EXPAND:
            if self.mask_zero:
                output_mask = K.not_equal(inputs, self.mask_zero)
            else:
                output_mask = None
        else:
            output_mask = mask
        return output_mask

    def compute_output_shape(self, input_shape):
        if self.mode == self.MODE_EXPAND:
            return input_shape + (self.output_dim,)
        if self.mode == self.MODE_CONCAT:
            return input_shape[:-1] + (input_shape[-1] + self.output_dim,)
        return input_shape

    def call(self, inputs, **kwargs):
        if self.mode == self.MODE_EXPAND:
            if K.dtype(inputs) != 'int32':
                inputs = K.cast(inputs, 'int32')
            return K.gather(
                self.embeddings,
                K.minimum(K.maximum(inputs, -self.input_dim), self.input_dim) + self.input_dim,
            )
        input_shape = K.shape(inputs)
        if self.mode == self.MODE_ADD:
            batch_size, seq_len, output_dim = input_shape[0], input_shape[1], input_shape[2]
        else:
            batch_size, seq_len, output_dim = input_shape[0], input_shape[1], self.output_dim
        pos_embeddings = K.tile(
            K.expand_dims(self.embeddings[:seq_len, :self.output_dim], axis=0),
            [batch_size, 1, 1],
        )
        if self.mode == self.MODE_ADD:
            return inputs + pos_embeddings
        return K.concatenate([inputs, pos_embeddings], axis=-1)


class TrigPosEmbedding(Layer):
    """Position embedding use sine and cosine functions.

    See: https://arxiv.org/pdf/1706.03762

    Expand mode:
        # Input shape
            2D tensor with shape: `(batch_size, sequence_length)`.

        # Output shape
            3D tensor with shape: `(batch_size, sequence_length, output_dim)`.

    Add mode:
        # Input shape
            3D tensor with shape: `(batch_size, sequence_length, feature_dim)`.

        # Output shape
            3D tensor with shape: `(batch_size, sequence_length, feature_dim)`.

    Concat mode:
        # Input shape
            3D tensor with shape: `(batch_size, sequence_length, feature_dim)`.

        # Output shape
            3D tensor with shape: `(batch_size, sequence_length, feature_dim + output_dim)`.
    """
    MODE_EXPAND = 'expand'
    MODE_ADD = 'add'
    MODE_CONCAT = 'concat'

    def __init__(self, mode=MODE_ADD, output_dim=None, **kwargs):
        """
        :param output_dim: The embedding dimension.
        :param kwargs:
        """
        if mode in [self.MODE_EXPAND, self.MODE_CONCAT]:
            if output_dim is None:
                raise NotImplementedError('`output_dim` is required in `%s` mode' % mode)
            if output_dim % 2 != 0:
                raise NotImplementedError('It does not make sense to use an odd output dimension: %d' % output_dim)
        self.mode = mode
        self.output_dim = output_dim
        self.supports_masking = True
        super(TrigPosEmbedding, self).__init__(**kwargs)

    def get_config(self):
        config = {
            'mode': self.mode,
            'output_dim': self.output_dim,
        }
        base_config = super(TrigPosEmbedding, self).get_config()
        config.update(base_config)
        return config

    def compute_mask(self, inputs, mask=None):
        return mask

    def compute_output_shape(self, input_shape):
        if self.mode == self.MODE_EXPAND:
            return input_shape + (self.output_dim,)
        if self.mode == self.MODE_CONCAT:
            return input_shape[:-1] + (input_shape[-1] + self.output_dim,)
        return input_shape

    def call(self, inputs, mask=None):
        input_shape = K.shape(inputs)
        if self.mode == self.MODE_ADD:
            batch_size, seq_len, output_dim = input_shape[0], input_shape[1], input_shape[2]
            pos_input = K.tile(K.expand_dims(K.arange(0, seq_len), axis=0), [batch_size, 1])
        elif self.mode == self.MODE_CONCAT:
            batch_size, seq_len, output_dim = input_shape[0], input_shape[1], self.output_dim
            pos_input = K.tile(K.expand_dims(K.arange(0, seq_len), axis=0), [batch_size, 1])
        else:
            output_dim = self.output_dim
            pos_input = inputs
        if K.dtype(pos_input) != K.floatx():
            pos_input = K.cast(pos_input, K.floatx())
        evens = K.arange(0, output_dim // 2) * 2
        odds = K.arange(0, output_dim // 2) * 2 + 1
        even_embd = K.sin(
            K.dot(
                K.expand_dims(pos_input, -1),
                K.expand_dims(1.0 / K.pow(10000.0, K.cast(evens, K.floatx()) / K.cast(output_dim, K.floatx())), 0)
            )
        )
        odd_embd = K.cos(
            K.dot(
                K.expand_dims(pos_input, -1),
                K.expand_dims(1.0 / K.pow(10000.0, K.cast((odds - 1), K.floatx()) / K.cast(output_dim, K.floatx())), 0)
            )
        )
        embd = K.stack([even_embd, odd_embd], axis=-1)
        output = K.reshape(embd, [-1, K.shape(inputs)[1], output_dim])
        if self.mode == self.MODE_CONCAT:
            output = K.concatenate([inputs, output], axis=-1)
        if self.mode == self.MODE_ADD:
            output += inputs
        return output


class TaskEmbedding(Layer):
    """Embedding for tasks.

        # Arguments
            input_dim: int > 0. Number of the tasks. Plus 1 if `mask_zero` is enabled.
            output_dim: int >= 0. Dimension of the dense embedding.
            embeddings_initializer: Initializer for the `embeddings` matrix.
            embeddings_regularizer: Regularizer function applied to the `embeddings` matrix.
            embeddings_constraint: Constraint function applied to the `embeddings` matrix.
            mask_zero: Generate zeros for 0 index if it is `True`.

        # Input shape
            Previous embeddings, 3D tensor with shape: `(batch_size, sequence_length, output_dim)`.
            Task IDs, 2D tensor with shape: `(batch_size, 1)`.

        # Output shape
            3D tensor with shape: `(batch_size, sequence_length, output_dim)`.
        """

    def __init__(self,
                 input_dim,
                 output_dim,
                 embeddings_initializer='uniform',
                 embeddings_regularizer=None,
                 embeddings_constraint=None,
                 mask_zero=False,
                 **kwargs):
        super(TaskEmbedding, self).__init__(**kwargs)
        self.supports_masking = True
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.embeddings_initializer = keras.initializers.get(embeddings_initializer)
        self.embeddings_regularizer = keras.regularizers.get(embeddings_regularizer)
        self.embeddings_constraint = keras.constraints.get(embeddings_constraint)
        self.mask_zero = mask_zero

        self.embeddings = None

    def build(self, input_shape):
        self.embeddings = self.add_weight(
            shape=(self.input_dim, self.output_dim),
            initializer=self.embeddings_initializer,
            regularizer=self.embeddings_regularizer,
            constraint=self.embeddings_constraint,
            name='embeddings',
        )
        super(TaskEmbedding, self).build(input_shape)

    def compute_mask(self, inputs, mask=None):
        output_mask = None
        if mask is not None:
            output_mask = mask[0]
        return output_mask

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def call(self, inputs, **kwargs):
        inputs, tasks = inputs
        if K.dtype(tasks) != 'int32':
            tasks = K.cast(tasks, 'int32')
        task_embed = K.gather(self.embeddings, tasks)
        if self.mask_zero:
            task_embed = task_embed * K.expand_dims(K.cast(K.not_equal(tasks, 0), K.floatx()), axis=-1)
        if K.backend() == 'theano':
            task_embed = K.tile(task_embed, (1, K.shape(inputs)[1], 1))
        return inputs + task_embed

    def get_config(self):
        config = {
            'input_dim': self.input_dim,
            'output_dim': self.output_dim,
            'embeddings_initializer': keras.initializers.serialize(self.embeddings_initializer),
            'embeddings_regularizer': keras.regularizers.serialize(self.embeddings_regularizer),
            'embeddings_constraint': keras.constraints.serialize(self.embeddings_constraint),
            'mask_zero': self.mask_zero,
        }
        base_config = super(TaskEmbedding, self).get_config()
        base_config.update(config)
        return base_config


def get_embedding(inputs, token_num, pos_num, embed_dim, dropout_rate=0.1, trainable=True):
    """Get embedding layer.

    See: https://arxiv.org/pdf/1810.04805.pdf

    :param inputs: Input layers.
    :param token_num: Number of tokens.
    :param pos_num: Maximum position.
    :param embed_dim: The dimension of all embedding layers.
    :param dropout_rate: Dropout rate.
    :param trainable: Whether the layers are trainable.
    :return: The merged embedding layer and weights of token embedding.
    """

    embeddings = [TokenEmbedding(input_dim=token_num,
                                 output_dim=embed_dim,
                                 mask_zero=True,
                                 trainable=trainable,
                                 name='Embedding-Token',
                                 )(inputs[0]),
                  Embedding(input_dim=2,
                            output_dim=embed_dim,
                            trainable=trainable,
                            name='Embedding-Segment',
                            )(inputs[1]),
                  ]
    embeddings[0], embed_weights = embeddings[0]
    embed_layer = Add(name='Embedding-Token-Segment')(embeddings)
    embed_layer = PositionEmbedding(input_dim=pos_num,
                                    output_dim=embed_dim,
                                    mode=PositionEmbedding.MODE_ADD,
                                    trainable=trainable,
                                    name='Embedding-Position',
                                    )(embed_layer)
    return embed_layer, embed_weights


def get_custom_objects():
    return {
        'EmbeddingRet': EmbeddingRet,
        'EmbeddingSim': EmbeddingSim,
        'PositionEmbedding': PositionEmbedding,
        'TrigPosEmbedding': TrigPosEmbedding,
    }



import tensorflow as tf
from tensorflow.keras import activations
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer
from tensorflow.keras.regularizers import l2


class FactorizationMachine(Layer):

    def __init__(self,
                 output_dim=1,
                 factor_dim=10,
                 name='FM',
                 activation=None,
                 use_weight=True,
                 use_bias=True,
                 kernel_reg=1e-4,
                 weight_reg=1e-4,
                 *args,
                 **kwargs):
        """
        Factorization Machines
        :param output_dim: 输出维度
        :param factor_dim: 隐含向量维度
        :param w_reg: the regularization coefficient of parameter w
        :param v_reg: the regularization coefficient of parameter v
        """
        super(FactorizationMachine, self).__init__(name=name, *args, **kwargs)
        self.output_dim = output_dim
        self.factor_dim = factor_dim
        self.activate = activation
        self.use_weight = use_weight
        self.use_bias = use_bias
        self.kernel_reg = kernel_reg
        self.weight_reg = weight_reg
        self.weight = self.bias = self.kernel = None
        self.activate_layer = None

    def build(self, input_shape):
        self.kernel = self.add_weight(name='kernel',
                                      shape=(input_shape[1], self.factor_dim),
                                      initializer='glorot_uniform',
                                      regularizer=l2(self.kernel_reg),
                                      trainable=True)
        if self.use_weight:
            self.weight = self.add_weight(name='weight',
                                          shape=(
                                              input_shape[1], self.output_dim),
                                          initializer='glorot_uniform',
                                          regularizer=l2(self.weight_reg),
                                          trainable=True)
        if self.use_bias:
            self.bias = self.add_weight(name='bias',
                                        shape=(self.output_dim,),
                                        initializer='zeros',
                                        trainable=True)

        self.activate_layer = activations.get(self.activate)
        super(FactorizationMachine, self).build(input_shape)

    def call(self, inputs, **kwargs):
        xv_a = K.square(K.dot(inputs, self.kernel))
        xv_b = K.dot(K.square(inputs), K.square(self.kernel))
        p = 0.5 * K.sum(xv_a - xv_b, 1)
        xv = K.repeat_elements(K.reshape(p, (-1, 1)), self.output_dim, axis=-1)

        res = xv
        if self.use_weight:
            res = res + K.dot(inputs, self.weight)
        if self.use_bias:
            res = res + self.bias

        output = K.reshape(res, (-1, self.output_dim))

        if self.activate_layer is not None:
            output = self.activate_layer(output)

        return output

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) == 2
        return input_shape[0], self.output_dim

    def get_config(self):
        config = {
            "output_dim": self.output_dim,
            "factor_dim": self.factor_dim,
            "name": self.name,
            "kernel_reg": self.kernel_reg,
            "weight_reg": self.weight_reg,
            "use_weight": self.use_weight,
            "use_bias": self.use_bias,
            "activation": self.activate
        }
        config.update(super(FactorizationMachine, self).get_config())
        return config


FM = FactorizationMachine


class FFM(Layer):
    def __init__(self,
                 field_num=3,
                 factor_dim=10,
                 use_weight=True,
                 use_bias=True,
                 weight_reg=1e-3,
                 kernel_reg=5e-2,
                 activation=None):
        """
        :param factor_dim:
        :param field_num:
        :param weight_reg:
        :param kernel_reg:
        """
        super(FFM, self).__init__()
        self.field_num = field_num
        self.factor_dim = factor_dim

        self.use_weight = use_weight
        self.use_bias = use_bias

        self.weight_reg = weight_reg
        self.kernel_reg = kernel_reg
        self.activation = activation

        self.kernel = self.weight = self.bias = self.activate_layer = None

    def build(self, input_shape):
        self.kernel = self.add_weight(name='kernel',
                                      shape=(
                                          input_shape[1], self.field_num, self.factor_dim),
                                      initializer=tf.random_normal_initializer(),
                                      regularizer=l2(self.kernel_reg),
                                      trainable=True)

        self.weight = self.add_weight(name='weight', shape=(input_shape[1], 1),
                                      initializer=tf.random_normal_initializer(),
                                      regularizer=l2(self.weight_reg),
                                      trainable=True)

        self.bias = self.add_weight(name='bias', shape=(
            1,), initializer=tf.zeros_initializer(), trainable=True)
        self.activate_layer = activations.get(self.activation)

    def call(self, inputs, **kwargs):
        field_f = tf.tensordot(inputs, self.kernel, axes=[1, 0])

        output = 0
        for i in range(self.field_num):
            for j in range(i + 1, self.field_num):
                output += tf.reduce_sum(tf.multiply(
                    field_f[:, i], field_f[:, j]), axis=1, keepdims=True)

        if self.use_weight:
            output += tf.matmul(inputs, self.weight)

        if self.use_bias:
            output += self.bias

        if self.activate_layer:
            output = self.activate_layer(output)

        return output

    def get_config(self):
        config = {
            "name": self.name,
            "field_num": self.field_num,
            "factor_dim": self.factor_dim,
            "use_weight": self.use_weight,
            "use_bias": self.use_bias,
            "weight_reg": self.weight_reg,
            "kernel_reg": self.kernel_reg,
            "activation": self.activation
        }
        config.update(super(FFM, self).get_config())
        return config



import efficientnet.keras as efn
import tensorflow.keras.backend as K
from keras.utils import get_file
from tensorflow import keras

from . import Backbone
from .retinanet import RetinaNetModel


class EfficientNetBackbone(Backbone):
    """ Describes backbone information and provides utility functions.
    """

    def __init__(self, backbone):
        super(EfficientNetBackbone, self).__init__(backbone)
        self.preprocess_image_func = None

    def retinanet(self, *args, **kwargs):
        """ Returns a retinanet model using the correct backbone.
        """
        return effnet_retinanet(*args, backbone=self.backbone, **kwargs)

    def download_imagenet(self):
        """ Downloads ImageNet weights and returns path to weights file.
        """
        from efficientnet.weights import IMAGENET_WEIGHTS_PATH
        from efficientnet.weights import IMAGENET_WEIGHTS_HASHES

        model_name = 'efficientnet-b' + self.backbone[-1]
        file_name = model_name + '_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5'
        file_hash = IMAGENET_WEIGHTS_HASHES[model_name][1]
        weights_path = get_file(file_name, IMAGENET_WEIGHTS_PATH + file_name, cache_subdir='models',
                                file_hash=file_hash)
        return weights_path

    def validate(self):
        """ Checks whether the backbone string is correct.
        """
        allowed_backbones = ['EfficientNetB0', 'EfficientNetB1', 'EfficientNetB2', 'EfficientNetB3', 'EfficientNetB4',
                             'EfficientNetB5', 'EfficientNetB6', 'EfficientNetB7']
        backbone = self.backbone.split('_')[0]

        if backbone not in allowed_backbones:
            raise ValueError('Backbone (\'{}\') not in allowed backbones ({}).'.format(backbone, allowed_backbones))

    def preprocess_image(self, inputs):
        """ Takes as input an image and prepares it for being passed through the network.
        """
        return efn.preprocess_input(inputs)


def effnet_retinanet(num_classes, backbone='EfficientNetB0', inputs=None, modifier=None, **kwargs):
    """ Constructs a retinanet model using a resnet backbone.

    Args
        num_classes: Number of classes to predict.
        backbone: Which backbone to use (one of ('resnet50', 'resnet101', 'resnet152')).
        inputs: The inputs to the network (defaults to a Tensor of shape (None, None, 3)).
        modifier: A function handler which can modify the backbone before using it in retinanet (this can be used to freeze backbone layers for example).

    Returns
        RetinaNet model with a ResNet backbone.
    """
    # choose default input
    if inputs is None:
        if K.image_data_format() == 'channels_first':
            inputs = keras.layers.Input(shape=(3, None, None))
        else:
            # inputs = keras.layers.Input(shape=(224, 224, 3))
            inputs = keras.layers.Input(shape=(None, None, 3))

    # get last conv layer from the end of each block [28x28, 14x14, 7x7]
    if backbone == 'EfficientNetB0':
        model = efn.EfficientNetB0(input_tensor=inputs, include_top=False, weights=None)
    elif backbone == 'EfficientNetB1':
        model = efn.EfficientNetB1(input_tensor=inputs, include_top=False, weights=None)
    elif backbone == 'EfficientNetB2':
        model = efn.EfficientNetB2(input_tensor=inputs, include_top=False, weights=None)
    elif backbone == 'EfficientNetB3':
        model = efn.EfficientNetB3(input_tensor=inputs, include_top=False, weights=None)
    elif backbone == 'EfficientNetB4':
        model = efn.EfficientNetB4(input_tensor=inputs, include_top=False, weights=None)
    elif backbone == 'EfficientNetB5':
        model = efn.EfficientNetB5(input_tensor=inputs, include_top=False, weights=None)
    elif backbone == 'EfficientNetB6':
        model = efn.EfficientNetB6(input_tensor=inputs, include_top=False, weights=None)
    elif backbone == 'EfficientNetB7':
        model = efn.EfficientNetB7(input_tensor=inputs, include_top=False, weights=None)
    else:
        raise ValueError('Backbone (\'{}\') is invalid.'.format(backbone))

    layer_outputs = ['block4a_expand_activation', 'block6a_expand_activation', 'top_activation']

    layer_outputs = [
        model.get_layer(name=layer_outputs[0]).output,  # 28x28
        model.get_layer(name=layer_outputs[1]).output,  # 14x14
        model.get_layer(name=layer_outputs[2]).output,  # 7x7
    ]
    # create the densenet backbone
    model = keras.models.Model(inputs=inputs, outputs=layer_outputs, name=model.name)

    # invoke modifier if given
    if modifier:
        model = modifier(model)

    # create the full model
    return RetinaNetModel(inputs=inputs, num_classes=num_classes, backbone_layers=model.outputs, **kwargs)


def EfficientNetB0_retinanet(num_classes, inputs=None, **kwargs):
    return effnet_retinanet(num_classes=num_classes, backbone='EfficientNetB0', inputs=inputs, **kwargs)


def EfficientNetB1_retinanet(num_classes, inputs=None, **kwargs):
    return effnet_retinanet(num_classes=num_classes, backbone='EfficientNetB1', inputs=inputs, **kwargs)


def EfficientNetB2_retinanet(num_classes, inputs=None, **kwargs):
    return effnet_retinanet(num_classes=num_classes, backbone='EfficientNetB2', inputs=inputs, **kwargs)


def EfficientNetB3_retinanet(num_classes, inputs=None, **kwargs):
    return effnet_retinanet(num_classes=num_classes, backbone='EfficientNetB3', inputs=inputs, **kwargs)


def EfficientNetB4_retinanet(num_classes, inputs=None, **kwargs):
    return effnet_retinanet(num_classes=num_classes, backbone='EfficientNetB4', inputs=inputs, **kwargs)


def EfficientNetB5_retinanet(num_classes, inputs=None, **kwargs):
    return effnet_retinanet(num_classes=num_classes, backbone='EfficientNetB5', inputs=inputs, **kwargs)


def EfficientNetB6_retinanet(num_classes, inputs=None, **kwargs):
    return effnet_retinanet(num_classes=num_classes, backbone='EfficientNetB6', inputs=inputs, **kwargs)


def EfficientNetB7_retinanet(num_classes, inputs=None, **kwargs):
    return effnet_retinanet(num_classes=num_classes, backbone='EfficientNetB7', inputs=inputs, **kwargs)



import numpy as np
import tensorflow as tf
from scipy import sparse as sp
from tensorflow.keras import backend as K
from tensorflow.python.ops import gen_sparse_ops
from tensorflow.python.ops.linalg.sparse import sparse as tfsp

SINGLE = 1  # Single mode    rank(x) = 2, rank(a) = 2
DISJOINT = SINGLE  # Disjoint mode  rank(x) = 2, rank(a) = 2
BATCH = 3  # Batch mode     rank(x) = 3, rank(a) = 3
MIXED = 4  # Mixed mode     rank(x) = 3, rank(a) = 2


def transpose(a, perm=None, name=None):
    """
    Transposes a according to perm, dealing automatically with sparsity.
    :param a: Tensor or SparseTensor with rank k.
    :param perm: permutation indices of size k.
    :param name: name for the operation.
    :return: Tensor or SparseTensor with rank k.
    """
    if K.is_sparse(a):
        transpose_op = tf.sparse.transpose
    else:
        transpose_op = tf.transpose

    if perm is None:
        perm = (1, 0)  # Make explicit so that shape will always be preserved
    return transpose_op(a, perm=perm, name=name)


def reshape(a, shape=None, name=None):
    """
    Reshapes a according to shape, dealing automatically with sparsity.
    :param a: Tensor or SparseTensor.
    :param shape: new shape.
    :param name: name for the operation.
    :return: Tensor or SparseTensor.
    """
    if K.is_sparse(a):
        reshape_op = tf.sparse.reshape
    else:
        reshape_op = tf.reshape

    return reshape_op(a, shape=shape, name=name)


def repeat(x, repeats):
    """
    Repeats elements of a Tensor (equivalent to np.repeat, but only for 1D
    tensors).
    :param x: rank 1 Tensor;
    :param repeats: rank 1 Tensor with same shape as x, the number of
    repetitions for each element;
    :return: rank 1 Tensor, of shape `(sum(repeats), )`.
    """
    x = tf.expand_dims(x, 1)
    max_repeats = tf.reduce_max(repeats)
    tile_repeats = [1, max_repeats]
    arr_tiled = tf.tile(x, tile_repeats)
    mask = tf.less(tf.range(max_repeats), tf.expand_dims(repeats, 1))
    result = tf.reshape(tf.boolean_mask(arr_tiled, mask), [-1])
    return result


def segment_top_k(x, I, ratio):
    """
    Returns indices to get the top K values in x segment-wise, according to
    the segments defined in I. K is not fixed, but it is defined as a ratio of
    the number of elements in each segment.
    :param x: a rank 1 Tensor;
    :param I: a rank 1 Tensor with segment IDs for x;
    :param ratio: float, ratio of elements to keep for each segment;
    :return: a rank 1 Tensor containing the indices to get the top K values of
    each segment in x.
    """
    rt = tf.RaggedTensor.from_value_rowids(x, I)
    row_lengths = rt.row_lengths()
    dense = rt.to_tensor(default_value=-np.inf)
    indices = tf.cast(tf.argsort(dense, direction="DESCENDING"), tf.int64)
    row_starts = tf.cast(rt.row_starts(), tf.int64)
    indices = indices + tf.expand_dims(row_starts, 1)
    row_lengths = tf.cast(
        tf.math.ceil(ratio * tf.cast(row_lengths, tf.float32)), tf.int32
    )
    return tf.RaggedTensor.from_tensor(indices, row_lengths).values


def indices_to_mask(indices, shape, dtype=tf.bool):
    """
    Return mask with true values at indices of the given shape.
    This can be used as an inverse to tf.where.
    :param indices: [nnz, k] or [nnz] Tensor indices of True values.
    :param shape: [k] or [] (scalar) Tensor shape/size of output.
    :return: Tensor of given shape and dtype.
    """
    indices = tf.convert_to_tensor(indices, dtype_hint=tf.int64)
    if indices.shape.ndims == 1:
        assert isinstance(shape, int) or shape.shape.ndims == 0
        indices = tf.expand_dims(indices, axis=1)
        if isinstance(shape, int):
            shape = tf.TensorShape([shape])
        else:
            shape = tf.expand_dims(shape, axis=0)
    else:
        indices.shape.assert_has_rank(2)
    assert indices.dtype.is_integer
    nnz = tf.shape(indices)[0]
    indices = tf.cast(indices, tf.int64)
    shape = tf.cast(shape, tf.int64)
    return tf.scatter_nd(indices, tf.ones((nnz,), dtype=dtype), shape)


def disjoint_signal_to_batch(X, I):
    """
    Converts a disjoint graph signal to batch node by zero-padding.

    :param X: Tensor, node features of shape (nodes, features).
    :param I: Tensor, graph IDs of shape `(n_nodes, )`;
    :return batch: Tensor, batched node features of shape (batch, N_max, n_node_features)
    """
    I = tf.cast(I, tf.int32)
    num_nodes = tf.math.segment_sum(tf.ones_like(I), I)
    start_index = tf.cumsum(num_nodes, exclusive=True)
    n_graphs = tf.shape(num_nodes)[0]
    max_n_nodes = tf.reduce_max(num_nodes)
    batch_n_nodes = tf.shape(I)[0]
    feature_dim = tf.shape(X)[-1]

    index = tf.range(batch_n_nodes)
    index = (index - tf.gather(start_index, I)) + (I * max_n_nodes)
    dense = tf.zeros((n_graphs * max_n_nodes, feature_dim), dtype=X.dtype)
    dense = tf.tensor_scatter_nd_update(dense, index[..., None], X)

    batch = tf.reshape(dense, (n_graphs, max_n_nodes, feature_dim))

    return batch


def _vectorised_get_cum_graph_size(nodes, graph_sizes):
    """
    Takes a list of node ids and graph sizes ordered by segment ID and returns the
    number of nodes contained in graphs with smaller segment ID.

    :param nodes: List of node ids of shape (nodes)
    :param graph_sizes: List of graph sizes (i.e. tf.math.segment_sum(tf.ones_like(I), I) where I are the
    segment IDs).
    :return: A list of shape (nodes) where each entry corresponds to the number of nodes contained in graphs
    with smaller segment ID for each node.
    """

    def get_cum_graph_size(node):
        cum_graph_sizes = tf.cumsum(graph_sizes, exclusive=True)
        indicator_if_smaller = tf.cast(node - cum_graph_sizes >= 0, tf.int32)
        graph_id = tf.reduce_sum(indicator_if_smaller) - 1
        return tf.cumsum(graph_sizes, exclusive=True)[graph_id]

    return tf.map_fn(get_cum_graph_size, nodes)


def disjoint_adjacency_to_batch(A, I):
    """
    Converts a disjoint adjacency matrix to batch node by zero-padding.

    :param A: Tensor, binary adjacency matrix of shape `(n_nodes, n_nodes)`;
    :param I: Tensor, graph IDs of shape `(n_nodes, )`;
    :return: Tensor, batched adjacency matrix of shape `(batch, N_max, N_max)`;
    """
    I = tf.cast(I, tf.int64)
    A = tf.cast(A, tf.float32)
    indices = A.indices
    values = tf.cast(A.values, tf.int64)
    i_nodes, j_nodes = indices[:, 0], indices[:, 1]

    graph_sizes = tf.math.segment_sum(tf.ones_like(I), I)
    max_n_nodes = tf.reduce_max(graph_sizes)
    n_graphs = tf.shape(graph_sizes)[0]
    relative_j_nodes = j_nodes - \
        _vectorised_get_cum_graph_size(j_nodes, graph_sizes)
    relative_i_nodes = i_nodes - \
        _vectorised_get_cum_graph_size(i_nodes, graph_sizes)
    spaced_i_nodes = I * max_n_nodes + relative_i_nodes
    new_indices = tf.transpose(tf.stack([spaced_i_nodes, relative_j_nodes]))

    new_indices = tf.cast(new_indices, tf.int32)
    n_graphs = tf.cast(n_graphs, tf.int32)
    max_n_nodes = tf.cast(max_n_nodes, tf.int32)

    dense_adjacency = tf.scatter_nd(
        new_indices, values, (n_graphs * max_n_nodes, max_n_nodes)
    )
    batch = tf.reshape(dense_adjacency, (n_graphs, max_n_nodes, max_n_nodes))
    batch = tf.cast(batch, tf.float32)
    return batch


def autodetect_mode(x, a):
    """
    Returns a code that identifies the data mode from the given node features
    and adjacency matrix(s).
    The output of this function can be used as follows:

    ```py
    from spektral.layers.ops import modes
    mode = modes.autodetect_mode(x, a)
    if mode == modes.SINGLE:
        print('Single!')
    elif mode == modes.BATCH:
        print('Batch!')
    elif mode == modes.MIXED:
        print('Mixed!')
    ```

    :param x: Tensor or SparseTensor representing the node features
    :param a: Tensor or SparseTensor representing the adjacency matrix(s)
    :return: mode of operation as an integer code.
    """
    x_ndim = K.ndim(x)
    a_ndim = K.ndim(a)
    if x_ndim == 2 and a_ndim == 2:
        return SINGLE
    elif x_ndim == 3 and a_ndim == 3:
        return BATCH
    elif x_ndim == 3 and a_ndim == 2:
        return MIXED
    else:
        raise ValueError(
            "Unknown mode for inputs x, a with ranks {} and {}"
            "respectively.".format(x_ndim, a_ndim)
        )


def dot(a, b):
    """
    Computes a @ b, for a, b of the same rank (both 2 or both 3).

    If the rank is 2, then the innermost dimension of `a` must match the
    outermost dimension of `b`.
    If the rank is 3, the first dimension of `a` and `b` must be equal and the
    function computes a batch matmul.

    Supports both dense and sparse multiplication (including sparse-sparse).

    :param a: Tensor or SparseTensor with rank 2 or 3.
    :param b: Tensor or SparseTensor with same rank as b.
    :return: Tensor or SparseTensor with rank 2 or 3.
    """
    a_ndim = K.ndim(a)
    b_ndim = K.ndim(b)
    assert a_ndim == b_ndim, "Expected equal ranks, got {} and {}" "".format(
        a_ndim, b_ndim
    )
    a_is_sparse = K.is_sparse(a)
    b_is_sparse = K.is_sparse(b)

    # Handle cases: rank 2 sparse-dense, rank 2 dense-sparse
    # In these cases we can use the faster sparse-dense matmul of tf.sparse
    if a_ndim == 2:
        if a_is_sparse and not b_is_sparse:
            return tf.sparse.sparse_dense_matmul(a, b)
        if not a_is_sparse and b_is_sparse:
            return transpose(tf.sparse.sparse_dense_matmul(transpose(b), transpose(a)))

    # Handle cases: rank 2 sparse-sparse, rank 3 sparse-dense,
    # rank 3 dense-sparse, rank 3 sparse-sparse
    # In these cases we can use the tfsp.CSRSparseMatrix implementation (slower,
    # but saves memory)
    if a_is_sparse:
        a = tfsp.CSRSparseMatrix(a)
    if b_is_sparse:
        b = tfsp.CSRSparseMatrix(b)
    if a_is_sparse or b_is_sparse:
        out = tfsp.matmul(a, b)
        if hasattr(out, "to_sparse_tensor"):
            return out.to_sparse_tensor()
        else:
            return out

    # Handle case: rank 2 dense-dense, rank 3 dense-dense
    # Here we use the standard dense operation
    return tf.matmul(a, b)


def mixed_mode_dot(a, b):
    """
    Computes the equivalent of `tf.einsum('ij,bjk->bik', a, b)`, but
    works for both dense and sparse inputs.

    :param a: Tensor or SparseTensor with rank 2.
    :param b: Tensor or SparseTensor with rank 3.
    :return: Tensor or SparseTensor with rank 3.
    """
    shp = K.int_shape(b)
    b_t = transpose(b, (1, 2, 0))
    b_t = reshape(b_t, (shp[1], -1))
    output = dot(a, b_t)
    output = reshape(output, (shp[1], shp[2], -1))
    output = transpose(output, (2, 0, 1))

    return output


def modal_dot(a, b, transpose_a=False, transpose_b=False):
    """
    Computes the matrix multiplication of a and b, handling the data modes
    automatically.

    This is a wrapper to standard matmul operations, for a and b with rank 2
    or 3, that:

    - Supports automatic broadcasting of the "batch" dimension if the two inputs
    have different ranks.
    - Supports any combination of dense and sparse inputs.

    This op is useful for multiplying matrices that represent batches of graphs
    in the different modes, for which the adjacency matrices may or may not be
    sparse and have different ranks from the node attributes.

    Additionally, it can also support the case where we have many adjacency
    matrices and only one graph signal (which is uncommon, but may still happen).

    If you know a-priori the type and shape of the inputs, it may be faster to
    use the built-in functions of TensorFlow directly instead.

    Examples:

        - `a` rank 2, `b` rank 2 -> `a @ b`
        - `a` rank 3, `b` rank 3 -> `[a[i] @ b[i] for i in range(len(a))]`
        - `a` rank 2, `b` rank 3 -> `[a @ b[i] for i in range(len(b))]`
        - `a` rank 3, `b` rank 2 -> `[a[i] @ b for i in range(len(a))]`

    :param a: Tensor or SparseTensor with rank 2 or 3;
    :param b: Tensor or SparseTensor with rank 2 or 3;
    :param transpose_a: transpose the innermost 2 dimensions of `a`;
    :param transpose_b: transpose the innermost 2 dimensions of `b`;
    :return: Tensor or SparseTensor with rank = max(rank(a), rank(b)).
    """
    a_ndim = K.ndim(a)
    b_ndim = K.ndim(b)
    assert a_ndim in (2, 3), "Expected a of rank 2 or 3, got {}".format(a_ndim)
    assert b_ndim in (2, 3), "Expected b of rank 2 or 3, got {}".format(b_ndim)

    if transpose_a:
        perm = None if a_ndim == 2 else (0, 2, 1)
        a = transpose(a, perm)
    if transpose_b:
        perm = None if b_ndim == 2 else (0, 2, 1)
        b = transpose(b, perm)

    if a_ndim == b_ndim:
        # ...ij,...jk->...ik
        return dot(a, b)
    elif a_ndim == 2:
        # ij,bjk->bik
        return mixed_mode_dot(a, b)
    else:  # a_ndim == 3
        # bij,jk->bik
        if not K.is_sparse(a) and not K.is_sparse(b):
            # Immediately fallback to standard dense matmul, no need to reshape
            return tf.matmul(a, b)

        # If either input is sparse, we use dot(a, b)
        # This implementation is faster than using rank 3 sparse matmul with tfsp
        a_shape = tf.shape(a)
        b_shape = tf.shape(b)
        a_flat = reshape(a, (-1, a_shape[2]))
        output = dot(a_flat, b)
        return reshape(output, (-1, a_shape[1], b_shape[1]))


def matmul_at_b_a(a, b):
    """
    Computes a.T @ b @ a, for a, b with rank 2 or 3.

    Supports automatic broadcasting of the "batch" dimension if the two inputs
    have different ranks.
    Supports any combination of dense and sparse inputs.

    :param a: Tensor or SparseTensor with rank 2 or 3.
    :param b: Tensor or SparseTensor with rank 2 or 3.
    :return: Tensor or SparseTensor with rank = max(rank(a), rank(b)).
    """
    at_b = modal_dot(a, b, transpose_a=True)
    at_b_a = modal_dot(at_b, a)

    return at_b_a


def matrix_power(a, k):
    """
    If a is a square matrix, computes a^k. If a is a rank 3 Tensor of square
    matrices, computes the exponent of each inner matrix.

    :param a: Tensor or SparseTensor with rank 2 or 3. The innermost two
    dimensions must be the same.
    :param k: int, the exponent to which to raise the matrices.
    :return: Tensor or SparseTensor with same rank as the input.
    """
    x_k = a
    for _ in range(k - 1):
        x_k = modal_dot(a, x_k)

    return x_k


def mixed_mode_support(scatter_fn):
    def _wrapper_mm_support(updates, indices, N):
        if len(tf.shape(updates)) == 3:
            updates = tf.transpose(updates, perm=(1, 0, 2))
        out = scatter_fn(updates, indices, N)
        if len(tf.shape(out)) == 3:
            out = tf.transpose(out, perm=(1, 0, 2))
        return out

    _wrapper_mm_support.__name__ = scatter_fn.__name__
    return _wrapper_mm_support


@mixed_mode_support
def scatter_sum(messages, indices, n_nodes):
    """
    Sums messages according to the segments defined by `indices`, with
    support for messages in single/disjoint mode (rank 2) and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be all zeros.
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_sum(messages, indices, n_nodes)


@mixed_mode_support
def scatter_mean(messages, indices, n_nodes):
    """
    Averages messages according to the segments defined by `indices`, with
    support for messages in single/disjoint mode (rank 2) and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be all zeros.
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_mean(messages, indices, n_nodes)


@mixed_mode_support
def scatter_max(messages, indices, n_nodes):
    """
    Max-reduces messages according to the segments defined by `indices`, with
    support for messages in single/disjoint mode (rank 2) and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be the minimum possible
    value for the dtype of the messages
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_max(messages, indices, n_nodes)


@mixed_mode_support
def scatter_min(messages, indices, n_nodes):
    """
    Min-reduces messages according to the segments defined by `indices`, with
    support for messages in single/disjoint mode (rank 2) and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be the maximum possible
    value for the dtype of the messages.
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_min(messages, indices, n_nodes)


@mixed_mode_support
def scatter_prod(messages, indices, n_nodes):
    """
    Multiplies messages element-wise according to the segments defined by
    `indices`, with support for messages in single/disjoint mode (rank 2)
    and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be all ones.
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_prod(messages, indices, n_nodes)


OP_DICT = {
    "sum": scatter_sum,
    "mean": scatter_mean,
    "max": scatter_max,
    "min": scatter_min,
    "prod": scatter_prod,
}


def unsorted_segment_softmax(x, indices, n_nodes=None):
    """
    Applies softmax along the segments of a Tensor. This operator is similar
    to the tf.math.segment_* operators, which apply a certain reduction to the
    segments. In this case, the output tensor is not reduced and maintains the
    same shape as the input.
    :param x: a Tensor. The softmax is applied along the first dimension.
    :param indices: a Tensor, indices to the segments.
    :param n_nodes: the number of unique segments in the indices. If `None`,
    n_nodes is calculated as the maximum entry in the indices plus 1.
    :return: a Tensor with the same shape as the input.
    """
    n_nodes = tf.reduce_max(indices) + 1 if n_nodes is None else n_nodes
    e_x = tf.exp(
        x - tf.gather(tf.math.unsorted_segment_max(x,
                                                   indices, n_nodes), indices)
    )
    e_x /= tf.gather(
        tf.math.unsorted_segment_sum(e_x, indices, n_nodes) + 1e-9, indices
    )
    return e_x


def serialize_scatter(identifier):
    if identifier in OP_DICT:
        return identifier
    elif hasattr(identifier, "__name__"):
        for k, v in OP_DICT.items():
            if v.__name__ == identifier.__name__:
                return k
        return None


def deserialize_scatter(scatter):
    if isinstance(scatter, str):
        if scatter in OP_DICT:
            return OP_DICT[scatter]
        else:
            if callable(scatter):
                return scatter
            else:
                raise ValueError(
                    "scatter must be callable or string in: {}.".format(
                        list(OP_DICT.keys())
                    )
                )


def sp_matrix_to_sp_tensor(x):
    """
    Converts a Scipy sparse matrix to a SparseTensor.
    :param x: a Scipy sparse matrix.
    :return: a SparseTensor.
    """
    if len(x.shape) != 2:
        raise ValueError("x must have rank 2")
    row, col, values = sp.find(x)
    out = tf.SparseTensor(
        indices=np.array([row, col]).T, values=values, dense_shape=x.shape
    )
    return tf.sparse.reorder(out)


def sp_batch_to_sp_tensor(a_list):
    """
    Converts a list of Scipy sparse matrices to a rank 3 SparseTensor.
    :param a_list: list of Scipy sparse matrices with the same shape.
    :return: SparseTensor of rank 3.
    """
    tensor_data = []
    for i, a in enumerate(a_list):
        row, col, values = sp.find(a)
        batch = np.ones_like(col) * i
        tensor_data.append((values, batch, row, col))
    tensor_data = list(map(np.concatenate, zip(*tensor_data)))

    out = tf.SparseTensor(
        indices=np.array(tensor_data[1:]).T,
        values=tensor_data[0],
        dense_shape=(len(a_list),) + a_list[0].shape,
    )

    return out


def dense_to_sparse(x):
    """
    Converts a Tensor to a SparseTensor.
    :param x: a Tensor.
    :return: a SparseTensor.
    """
    indices = tf.where(tf.not_equal(x, 0))
    values = tf.gather_nd(x, indices)
    shape = tf.shape(x, out_type=tf.int64)
    return tf.SparseTensor(indices, values, shape)


def add_self_loops(a, fill=1.0):
    """
    Adds self-loops to the given adjacency matrix. Self-loops are added only for
    those node that don't have a self-loop already, and are assigned a weight
    of `fill`.
    :param a: a square SparseTensor.
    :param fill: the fill value for the new self-loops. It will be cast to the
    dtype of `a`.
    :return: a SparseTensor with the same shape as the input.
    """
    indices = a.indices
    values = a.values
    N = tf.shape(a, out_type=indices.dtype)[0]

    mask_od = indices[:, 0] != indices[:, 1]
    mask_sl = ~mask_od
    mask_od.set_shape([None])  # For compatibility with TF 2.2
    mask_sl.set_shape([None])

    indices_od = indices[mask_od]
    indices_sl = indices[mask_sl]

    values_sl = tf.fill((N,), tf.cast(fill, values.dtype))
    values_sl = tf.tensor_scatter_nd_update(
        values_sl, indices_sl[:, 0:1], values[mask_sl]
    )

    indices_sl = tf.range(N, dtype=indices.dtype)[:, None]
    indices_sl = tf.repeat(indices_sl, 2, -1)
    indices = tf.concat((indices_od, indices_sl), 0)

    values_od = values[mask_od]
    values = tf.concat((values_od, values_sl), 0)

    out = tf.SparseTensor(indices, values, (N, N))

    return tf.sparse.reorder(out)


def add_self_loops_indices(indices, n_nodes=None):
    """
    Given the indices of a square SparseTensor, adds the diagonal entries (i, i)
    and returns the reordered indices.
    :param indices: Tensor of rank 2, the indices to a SparseTensor.
    :param n_nodes: the size of the n_nodes x n_nodes SparseTensor indexed by
    the indices. If `None`, n_nodes is calculated as the maximum entry in the
    indices plus 1.
    :return: Tensor of rank 2, the indices to a SparseTensor.
    """
    n_nodes = tf.reduce_max(indices) + 1 if n_nodes is None else n_nodes
    row, col = indices[..., 0], indices[..., 1]
    mask = tf.ensure_shape(row != col, row.shape)
    sl_indices = tf.range(n_nodes, dtype=row.dtype)[:, None]
    sl_indices = tf.repeat(sl_indices, 2, -1)
    indices = tf.concat((indices[mask], sl_indices), 0)
    dummy_values = tf.ones_like(indices[:, 0])
    indices, _ = gen_sparse_ops.sparse_reorder(
        indices, dummy_values, (n_nodes, n_nodes)
    )
    return indices


def _square_size(dense_shape):
    dense_shape = tf.unstack(dense_shape)
    size = dense_shape[0]
    for d in dense_shape[1:]:
        tf.debugging.assert_equal(size, d)
    return d


def _indices_to_inverse_map(indices, size):
    """
    Compute inverse indices of a gather.
    :param indices: Tensor, forward indices, rank 1
    :param size: Tensor, size of pre-gathered input, rank 0
    :return: Tensor, inverse indices, shape [size]. Zero values everywhere
    except at indices.
    """
    indices = tf.cast(indices, tf.int64)
    size = tf.cast(size, tf.int64)
    return tf.scatter_nd(
        tf.expand_dims(indices, axis=-1),
        tf.range(tf.shape(indices, out_type=tf.int64)[0]),
        tf.expand_dims(size, axis=-1),
    )


def _boolean_mask_sparse(a, mask, axis, inverse_map, out_size):
    """
    SparseTensor equivalent to tf.boolean_mask.
    :param a: SparseTensor of rank k and nnz non-zeros.
    :param mask: rank-1 bool Tensor.
    :param axis: int, axis on which to mask. Must be in [-k, k).
    :param out_size: number of true entires in mask. Computed if not given.
    :return masked_a: SparseTensor masked along the given axis.
    :return values_mask: bool Tensor indicating surviving edges, shape [nnz].
    """
    mask = tf.convert_to_tensor(mask)
    values_mask = tf.gather(mask, a.indices[:, axis], axis=0)
    dense_shape = tf.tensor_scatter_nd_update(
        a.dense_shape, [[axis]], [out_size])
    indices = tf.boolean_mask(a.indices, values_mask)
    indices = tf.unstack(indices, axis=-1)
    indices[axis] = tf.gather(inverse_map, indices[axis])
    indices = tf.stack(indices, axis=-1)
    a = tf.SparseTensor(
        indices,
        tf.boolean_mask(a.values, values_mask),
        dense_shape,
    )
    return (a, values_mask)


def _boolean_mask_sparse_square(a, mask, inverse_map, out_size):
    """
    Apply boolean_mask to every axis of a SparseTensor.
    :param a: SparseTensor with uniform dimensions and nnz non-zeros.
    :param mask: boolean mask.
    :param inverse_map: Tensor of new indices, shape [nnz]. Computed if None.
    :out_size: number of True values in mask. Computed if None.
    :return a: SparseTensor with uniform dimensions.
    :return values_mask: bool Tensor of shape [nnz] indicating valid edges.
    """
    mask = tf.convert_to_tensor(mask)
    values_mask = tf.reduce_all(tf.gather(mask, a.indices, axis=0), axis=-1)
    dense_shape = [out_size] * a.shape.ndims
    indices = tf.boolean_mask(a.indices, values_mask)
    indices = tf.gather(inverse_map, indices)
    a = tf.SparseTensor(indices, tf.boolean_mask(
        a.values, values_mask), dense_shape)
    return (a, values_mask)


def boolean_mask_sparse(a, mask, axis=0):
    """
    SparseTensor equivalent to tf.boolean_mask.
    :param a: SparseTensor of rank k and nnz non-zeros.
    :param mask: rank-1 bool Tensor.
    :param axis: int, axis on which to mask. Must be in [-k, k).
    :return masked_a: SparseTensor masked along the given axis.
    :return values_mask: bool Tensor indicating surviving values, shape [nnz].
    """
    i = tf.squeeze(tf.where(mask), axis=1)
    out_size = tf.math.count_nonzero(mask)
    in_size = a.dense_shape[axis]
    inverse_map = _indices_to_inverse_map(i, in_size)
    return _boolean_mask_sparse(
        a, mask, axis=axis, inverse_map=inverse_map, out_size=out_size
    )


def boolean_mask_sparse_square(a, mask):
    """
    Apply mask to every axis of SparseTensor a.
    :param a: SparseTensor, square, nnz non-zeros.
    :param mask: boolean mask with size equal to each dimension of a.
    :return masked_a: SparseTensor
    :return values_mask: bool tensor of shape [nnz] indicating valid values.
    """
    i = tf.squeeze(tf.where(mask), axis=-1)
    out_size = tf.size(i)
    in_size = _square_size(a.dense_shape)
    inverse_map = _indices_to_inverse_map(i, in_size)
    return _boolean_mask_sparse_square(
        a, mask, inverse_map=inverse_map, out_size=out_size
    )


def gather_sparse(a, indices, axis=0, mask=None):
    """
    SparseTensor equivalent to tf.gather, assuming indices are sorted.
    :param a: SparseTensor of rank k and nnz non-zeros.
    :param indices: rank-1 int Tensor, rows or columns to keep.
    :param axis: int axis to apply gather to.
    :param mask: boolean mask corresponding to indices. Computed if not provided.
    :return gathered_a: SparseTensor masked along the given axis.
    :return values_mask: bool Tensor indicating surviving values, shape [nnz].
    """
    in_size = _square_size(a.dense_shape)
    out_size = tf.size(indices)
    if mask is None:
        mask = indices_to_mask(indices, in_size)
    inverse_map = _indices_to_inverse_map(indices, in_size)
    return _boolean_mask_sparse(
        a, mask, axis=axis, inverse_map=inverse_map, out_size=out_size
    )


def gather_sparse_square(a, indices, mask=None):
    """
    Gather on every axis of a SparseTensor.
    :param a: SparseTensor of rank k and nnz non-zeros.
    :param indices: rank-1 int Tensor, rows and columns to keep.
    :param mask: boolean mask corresponding to indices. Computed if not provided.
    :return gathered_a: SparseTensor of the gathered input.
    :return values_mask: bool Tensor indicating surviving values, shape [nnz].
    """
    in_size = _square_size(a.dense_shape)
    out_size = tf.size(indices)
    if mask is None:
        mask = indices_to_mask(indices, in_size)
    inverse_map = _indices_to_inverse_map(indices, in_size)
    return _boolean_mask_sparse_square(
        a, mask, inverse_map=inverse_map, out_size=out_size
    )


def normalize_A(A):
    """
    Computes symmetric normalization of A, dealing with sparse A and batch mode
    automatically.
    :param A: Tensor or SparseTensor with rank k = {2, 3}.
    :return: Tensor or SparseTensor of rank k.
    """
    D = degrees(A)
    D = tf.sqrt(D)[:, None] + K.epsilon()
    perm = (0, 2, 1) if K.ndim(A) == 3 else (1, 0)
    output = (A / D) / transpose(D, perm=perm)

    return output


def degrees(A):
    """
    Computes the degrees of each node in A, dealing with sparse A and batch mode
    automatically.
    :param A: Tensor or SparseTensor with rank k = {2, 3}.
    :return: Tensor or SparseTensor of rank k - 1.
    """
    if K.is_sparse(A):
        D = tf.sparse.reduce_sum(A, axis=-1)
    else:
        D = tf.reduce_sum(A, axis=-1)

    return D


def degree_matrix(A, return_sparse_batch=False):
    """
    Computes the degree matrix of A, deals with sparse A and batch mode
    automatically.
    :param A: Tensor or SparseTensor with rank k = {2, 3}.
    :param return_sparse_batch: if operating in batch mode, return a
    SparseTensor. Note that the sparse degree Tensor returned by this function
    cannot be used for sparse matrix multiplication afterwards.
    :return: SparseTensor of rank k.
    """
    D = degrees(A)

    batch_mode = K.ndim(D) == 2
    N = tf.shape(D)[-1]
    batch_size = tf.shape(D)[0] if batch_mode else 1

    inner_index = tf.tile(tf.stack([tf.range(N)] * 2, axis=1), (batch_size, 1))
    if batch_mode:
        if return_sparse_batch:
            outer_index = repeat(
                tf.range(batch_size), tf.ones(
                    batch_size) * tf.cast(N, tf.float32)
            )
            indices = tf.concat([outer_index[:, None], inner_index], 1)
            dense_shape = (batch_size, N, N)
        else:
            return tf.linalg.diag(D)
    else:
        indices = inner_index
        dense_shape = (N, N)

    indices = tf.cast(indices, tf.int64)
    values = tf.reshape(D, (-1,))
    return tf.SparseTensor(indices, values, dense_shape)



from notekeras.backend import backend as K
from notekeras.backend import keras, layers


class FeedForward(layers.Layer):
    """Position-wise feed-forward layer.

    # Arguments
        units: int >= 0. Dimension of hidden units.
        activation: Activation function to use
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        kernel_constraint: Constraint function applied to the `kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.
        dropout_rate: 0.0 <= float <= 1.0. Dropout rate for hidden units.

    # Input shape
        3D tensor with shape: `(batch_size, ..., input_dim)`.

    # Output shape
        3D tensor with shape: `(batch_size, ..., input_dim)`.

    # References
        - [Attention is All You Need](https://arxiv.org/pdf/1706.03762.pdf)
    """

    def __init__(self,
                 units,
                 activation='relu',
                 use_bias=True,
                 kernel_initializer='glorot_normal',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 dropout_rate=0.0,
                 **kwargs):
        self.supports_masking = True
        self.units = units
        self.activation = keras.activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.kernel_regularizer = keras.regularizers.get(kernel_regularizer)
        self.bias_regularizer = keras.regularizers.get(bias_regularizer)
        self.kernel_constraint = keras.constraints.get(kernel_constraint)
        self.bias_constraint = keras.constraints.get(bias_constraint)
        self.dropout_rate = dropout_rate
        self.W1, self.b1 = None, None
        self.W2, self.b2 = None, None
        super(FeedForward, self).__init__(**kwargs)

    def get_config(self):
        config = {
            'units': self.units,
            'activation': keras.activations.serialize(self.activation),
            'use_bias': self.use_bias,
            'kernel_initializer': keras.initializers.serialize(self.kernel_initializer),
            'bias_initializer': keras.initializers.serialize(self.bias_initializer),
            'kernel_regularizer': keras.regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer': keras.regularizers.serialize(self.bias_regularizer),
            'kernel_constraint': keras.constraints.serialize(self.kernel_constraint),
            'bias_constraint': keras.constraints.serialize(self.bias_constraint),
            'dropout_rate': self.dropout_rate,
        }
        base_config = super(FeedForward, self).get_config()
        # return dict(list(base_config.items()) + list(config.items()))
        return dict(base_config, **config)

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_mask(self, inputs, input_mask=None):
        return input_mask

    def build(self, input_shape):
        feature_dim = int(input_shape[-1])
        self.W1 = self.add_weight(shape=(feature_dim, self.units),
                                  initializer=self.kernel_initializer,
                                  regularizer=self.kernel_regularizer,
                                  constraint=self.kernel_constraint,
                                  name='{}_W1'.format(self.name),
                                  )
        if self.use_bias:
            self.b1 = self.add_weight(shape=(self.units,),
                                      initializer=self.bias_initializer,
                                      regularizer=self.bias_regularizer,
                                      constraint=self.bias_constraint,
                                      name='{}_b1'.format(self.name),
                                      )
        self.W2 = self.add_weight(shape=(self.units, feature_dim),
                                  initializer=self.kernel_initializer,
                                  regularizer=self.kernel_regularizer,
                                  constraint=self.kernel_constraint,
                                  name='{}_W2'.format(self.name),
                                  )
        if self.use_bias:
            self.b2 = self.add_weight(shape=(feature_dim,),
                                      initializer=self.bias_initializer,
                                      regularizer=self.bias_regularizer,
                                      constraint=self.bias_constraint,
                                      name='{}_b2'.format(self.name),
                                      )
        super(FeedForward, self).build(input_shape)

    def call(self, x, mask=None, training=None):
        h = K.dot(x, self.W1)
        if self.use_bias:
            h = K.bias_add(h, self.b1)
        if self.activation is not None:
            h = self.activation(h)
        if 0.0 < self.dropout_rate < 1.0:
            def dropped_inputs():
                return K.dropout(h, self.dropout_rate, K.shape(h))

            h = K.in_train_phase(dropped_inputs, h, training=training)
        y = K.dot(h, self.W2)
        if self.use_bias:
            y = K.bias_add(y, self.b2)
        return y



import logging

from tensorflow.keras import Input
from tensorflow.keras.layers import Conv1D, MaxPool1D, Dense, Flatten, concatenate, Embedding
from tensorflow.keras.models import Model
from tensorflow.keras.utils import plot_model


def textcnn(max_sequence_length, max_token_num, embedding_dim, output_dim, model_img_path=None, embedding_matrix=None):
    """ TextCNN: 1. embedding layers, 2.convolution layer, 3.max-pooling, 4.softmax layer. """
    x_input = Input(shape=(max_sequence_length,))
    logging.info("x_input.shape: %s" % str(x_input.shape))  # (?, 60)

    if embedding_matrix is None:
        x_emb = Embedding(input_dim=max_token_num, output_dim=embedding_dim, input_length=max_sequence_length)(x_input)
    else:
        x_emb = Embedding(input_dim=max_token_num, output_dim=embedding_dim, input_length=max_sequence_length,
                          weights=[embedding_matrix], trainable=True)(x_input)
    logging.info("x_emb.shape: %s" % str(x_emb.shape))  # (?, 60, 300)

    pool_output = []
    kernel_sizes = [2, 3, 4]
    for kernel_size in kernel_sizes:
        c = Conv1D(filters=2, kernel_size=kernel_size, strides=1)(x_emb)
        p = MaxPool1D(pool_size=int(c.shape[1]))(c)
        pool_output.append(p)
        logging.info("kernel_size: %s \t c.shape: %s \t p.shape: %s" % (kernel_size, str(c.shape), str(p.shape)))
    pool_output = concatenate([p for p in pool_output])
    logging.info("pool_output.shape: %s" % str(pool_output.shape))  # (?, 1, 6)

    x_flatten = Flatten()(pool_output)  # (?, 6)
    y = Dense(output_dim, activation='softmax')(x_flatten)  # (?, 2)
    logging.info("y.shape: %s \n" % str(y.shape))

    model = Model([x_input], outputs=[y])
    if model_img_path:
        plot_model(model, to_file=model_img_path, show_shapes=True, show_layer_names=False)
    model.summary()
    return model



def freeze(model):
    """ Set all layers in a model to non-trainable.

    The weights for these layers will not be updated during training.

    This function modifies the given model in-place,
    but it also returns the modified model to allow easy chaining with other functions.
    """
    for layer in model.layers:
        layer.trainable = False
    return model
